"""Dataset Metadata class."""

import copy
import json
import logging
import os
from collections import defaultdict

import numpy as np
import pandas as pd
from rdt import HyperTransformer, transformers

from rctgan.constraints import Constraint
from rctgan.metadata import visualization
from rctgan.metadata.errors import MetadataError

LOGGER = logging.getLogger(__name__)


def _read_csv_dtypes(table_meta):
    """Get the dtypes specification that needs to be passed to read_csv."""
    dtypes = dict()
    for name, field in table_meta['fields'].items():
        field_type = field['type']
        if field_type == 'id' and field.get('subtype', 'integer') == 'string':
            dtypes[name] = str

    return dtypes


def _parse_dtypes(data, table_meta):
    """Convert the data columns to the right dtype after loading the CSV."""
    for name, field in table_meta['fields'].items():
        field_type = field['type']
        if field_type == 'datetime':
            datetime_format = field.get('format')
            data[name] = pd.to_datetime(data[name], format=datetime_format, exact=False)
        elif field_type == 'numerical' and field.get('subtype') == 'integer':
            data[name] = data[name].dropna().astype(np.int64)
        elif field_type == 'id' and field.get('subtype', 'integer') == 'integer':
            data[name] = data[name].dropna().astype(np.int64)

    return data


def _load_csv(root_path, table_meta):
    """Load a CSV with the right dtypes and then parse the columns."""
    relative_path = os.path.join(root_path, table_meta['path'])
    dtypes = _read_csv_dtypes(table_meta)

    data = pd.read_csv(relative_path, dtype=dtypes)
    data = _parse_dtypes(data, table_meta)

    return data


class Metadata:
    """Dataset Metadata.

    The Metadata class provides a unified layer of abstraction over the dataset
    metadata, which includes both the necessary details to load the data from
    the hdd and to know how to parse and transform it to numerical data.

    Args:
        metadata (str or dict):
            Path to a ``json`` file that contains the metadata or a ``dict`` representation
            of ``metadata`` following the same structure.
        root_path (str):
            The path to which the paths found inside the metadata are relative.
            If not given, it defaults to the folder where the metadata is located, or
            to None, if no metadata is given.
    """

    _child_map = None
    _hyper_transformers = None
    _metadata = None
    _parent_map = None

    root_path = None

    _FIELD_TEMPLATES = {
        'i': {
            'type': 'numerical',
            'subtype': 'integer',
        },
        'f': {
            'type': 'numerical',
            'subtype': 'float',
        },
        'O': {
            'type': 'categorical',
        },
        'b': {
            'type': 'boolean',
        },
        'M': {
            'type': 'datetime',
        }
    }
    _DTYPES = {
        ('categorical', None): 'object',
        ('boolean', None): 'bool',
        ('numerical', None): 'float',
        ('numerical', 'float'): 'float',
        ('numerical', 'integer'): 'int',
        ('datetime', None): 'datetime64',
        ('id', None): 'int',
        ('id', 'integer'): 'int',
        ('id', 'string'): 'str'
    }

    def _analyze_relationships(self):
        """Extract information about child-parent relationships.

        Creates the following attributes:
            * ``_child_map``: set of child tables that each table has.
            * ``_parent_map``: set ot parents that each table has.
        """
        self._child_map = defaultdict(set)
        self._parent_map = defaultdict(set)

        for table, table_meta in self._metadata['tables'].items():
            if table_meta.get('use', True):
                for field_meta in table_meta['fields'].values():
                    ref = field_meta.get('ref')
                    if ref:
                        parent = ref['table']
                        self._child_map[parent].add(table)
                        self._parent_map[table].add(parent)

    @staticmethod
    def _dict_metadata(metadata):
        """Get a metadata ``dict`` with SDV format.

        For each table create a dict of fields from a previous list of fields.

        Args:
            metadata (dict):
                Original metadata to format.

        Returns:
            dict:
                Formated metadata dict.
        """
        new_metadata = copy.deepcopy(metadata)
        tables = new_metadata['tables']
        if isinstance(tables, dict):
            new_metadata['tables'] = {
                table: meta
                for table, meta in tables.items()
                if meta.pop('use', True)
            }
            return new_metadata

        new_tables = dict()
        for table in tables:
            if table.pop('use', True):
                new_tables[table.pop('name')] = table

                fields = table['fields']
                new_fields = dict()
                for field in fields:
                    new_fields[field.pop('name')] = field

                table['fields'] = new_fields

        new_metadata['tables'] = new_tables

        return new_metadata

    def __init__(self, metadata=None, root_path=None):
        if isinstance(metadata, str):
            self.root_path = root_path or os.path.dirname(metadata)
            with open(metadata) as metadata_file:
                metadata = json.load(metadata_file)
        else:
            self.root_path = root_path or '.'

        if metadata is not None:
            self._metadata = self._dict_metadata(metadata)
        else:
            self._metadata = {'tables': {}}

        self._hyper_transformers = dict()
        self._analyze_relationships()

    def get_children(self, table_name):
        """Get tables for which the given table is parent.

        Args:
            table_name (str):
                Name of the table from which to get the children.

        Returns:
            set:
                Set of children for the given table.
        """
        return self._child_map[table_name]

    def get_parents(self, table_name):
        """Get tables for with the given table is child.

        Args:
            table_name (str):
                Name of the table from which to get the parents.

        Returns:
            set:
                Set of parents for the given table.
        """
        return self._parent_map[table_name]

    def get_table_meta(self, table_name):
        """Get the metadata dict for a table.

        Args:
            table_name (str):
                Name of table to get data for.

        Returns:
            dict:
                table metadata

        Raises:
            ValueError:
                If table does not exist in this metadata.
        """
        table = self._metadata['tables'].get(table_name)
        if table is None:
            raise ValueError('Table "{}" does not exist'.format(table_name))

        return copy.deepcopy(table)

    def get_tables(self):
        """Get the list with the tables found in this metadata.

        Returns:
            list:
                table names.
        """
        return list(self._metadata['tables'].keys())

    def get_field_meta(self, table_name, field_name):
        """Get the metadata dict for a table.

        Args:
            table_name (str):
                Name of the table to which the field belongs.
            field_name (str):
                Name of the field to get data for.

        Returns:
            dict:
                field metadata

        Raises:
            ValueError:
                If the table or the field do not exist in this metadata.
        """
        field_meta = self.get_fields(table_name).get(field_name)
        if field_meta is None:
            raise ValueError(
                'Table "{}" does not contain a field name "{}"'.format(table_name, field_name))

        return copy.deepcopy(field_meta)

    def get_fields(self, table_name):
        """Get the metadata of the fields of the indicated table.

        Args:
            table_name (str):
                Name of the table to get the fields from.

        Returns:
            dict:
                Mapping of field names and their metadata dicts.

        Raises:
            ValueError:
                If table does not exist in this metadata.
        """
        return self.get_table_meta(table_name)['fields']

    def get_primary_key(self, table_name):
        """Get the name of the primary key of the indicated table, if it has one.

        Args:
            table_name (str):
                Name of table for which to get the primary key field.

        Returns:
            str or None:
                Primary key field name. ``None`` if the table has no primary key.

        Raises:
            ValueError:
                If table does not exist in this metadata.
        """
        return self.get_table_meta(table_name).get('primary_key')

    def get_foreign_keys(self, parent, child):
        """Get the name of all the fields in the child that are foreign keys to this parent.

        If there is no relationship between the two tables an empty list is returned.

        Args:
            parent (str):
                Name of the parent table.
            child (str):
                Name of the child table.

        Returns:
            list[str]:
                List of foreign key names.
        """
        foreign_keys = []
        for name, field in self.get_fields(child).items():
            ref = field.get('ref')
            if ref and ref['table'] == parent:
                foreign_keys.append(name)

        return foreign_keys

    def load_table(self, table_name):
        """Load the data of the indicated table as a DataFrame.

        Args:
            table_name (str):
                Name of the table to load.

        Returns:
            pandas.DataFrame:
                DataFrame with the contents of the table.

        Raises:
            ValueError:
                If table does not exist in this metadata.
        """
        LOGGER.info('Loading table %s', table_name)
        table_meta = self.get_table_meta(table_name)
        return _load_csv(self.root_path, table_meta)

    def load_tables(self, tables=None):
        """Get a dictionary with data from multiple tables.

        If a ``tables`` list is given, only load the indicated tables.
        Otherwise, load all the tables from this metadata.

        Args:
            tables (list):
                List of table names. Defaults to ``None``.

        Returns:
            dict(str, pandasd.DataFrame):
                mapping of table names and their data loaded as ``pandas.DataFrame`` instances.
        """
        return {
            table_name: self.load_table(table_name)
            for table_name in tables or self.get_tables()
        }

    def get_dtypes(self, table_name, ids=False, errors=None):
        """Get a ``dict`` with the ``dtypes`` for each field of a given table.

        Args:
            table_name (str):
                Table name for which to retrive the ``dtypes``.
            ids (bool):
                Whether or not include the id fields. Defaults to ``False``.

        Returns:
            dict:
                Dictionary that contains the field names and data types from a table.

        Raises:
            ValueError:
                If a field has an invalid type or subtype or if the table does not
                exist in this metadata.
        """
        errors = [] if errors is None else errors
        dtypes = dict()
        table_meta = self.get_table_meta(table_name)
        for name, field in table_meta['fields'].items():
            field_type = field['type']
            field_subtype = field.get('subtype')
            dtype = self._DTYPES.get((field_type, field_subtype))
            if not dtype:
                errors.append(
                    'Invalid type and subtype combination for field {}: ({}, {})'.format(
                        name, field_type, field_subtype)
                )
            else:
                if ids and field_type == 'id':
                    if (name != table_meta.get('primary_key')) and not field.get('ref'):
                        for child_table in self.get_children(table_name):
                            if name in self.get_foreign_keys(table_name, child_table):
                                break

                if ids or (field_type != 'id'):
                    dtypes[name] = dtype

        return dtypes

    def _get_pii_fields(self, table_name):
        """Get the ``pii_category`` for each field of the table that contains PII.

        Args:
            table_name (str):
                Table name for which to get the pii fields.

        Returns:
            dict:
                pii field names and categories.
        """
        pii_fields = dict()
        for name, field in self.get_table_meta(table_name)['fields'].items():
            if field['type'] == 'categorical' and field.get('pii', False):
                pii_fields[name] = field['pii_category']

        return pii_fields

    @staticmethod
    def _get_transformers(dtypes, pii_fields):
        """Create the transformer instances needed to process the given dtypes.

        Temporary drop-in replacement of ``HyperTransformer._analyze`` method,
        before RDT catches up.

        Args:
            dtypes (dict):
                mapping of field names and dtypes.
            pii_fields (dict):
                mapping of pii field names and categories.

        Returns:
            dict:
                mapping of field names and transformer instances.
        """
        transformers_dict = dict()
        for name, dtype in dtypes.items():
            dtype = np.dtype(dtype)
            if dtype.kind == 'i':
                transformer = transformers.NumericalTransformer(dtype=int)
            elif dtype.kind == 'f':
                transformer = transformers.NumericalTransformer(dtype=float)
            elif dtype.kind == 'O':
                anonymize = pii_fields.get(name)
                transformer = transformers.CategoricalTransformer(anonymize=anonymize)
            elif dtype.kind == 'b':
                transformer = transformers.BooleanTransformer()
            elif dtype.kind == 'M':
                transformer = transformers.DatetimeTransformer()
            else:
                raise ValueError('Unsupported dtype: {}'.format(dtype))

            LOGGER.info('Loading transformer %s for field %s',
                        transformer.__class__.__name__, name)
            transformers_dict[name] = transformer

        return transformers_dict

    def _load_hyper_transformer(self, table_name):
        """Create and return a new ``rdt.HyperTransformer`` instance for a table.

        First get the ``dtypes`` and ``pii fields`` from a given table, then use
        those to build a transformer dictionary to be used by the ``HyperTransformer``.

        Args:
            table_name (str):
                Table name for which to load the HyperTransformer.

        Returns:
            rdt.HyperTransformer:
                Instance of ``rdt.HyperTransformer`` for the given table.
        """
        dtypes = self.get_dtypes(table_name)
        pii_fields = self._get_pii_fields(table_name)
        transformers_dict = self._get_transformers(dtypes, pii_fields)
        return HyperTransformer(field_transformers=transformers_dict)

    def transform(self, table_name, data):
        """Transform data for a given table.

        If the ``HyperTransformer`` for a table is ``None`` it is created.

        Args:
            table_name (str):
                Name of the table that is being transformer.
            data (pandas.DataFrame):
                Table data.

        Returns:
            pandas.DataFrame:
                Transformed data.
        """
        hyper_transformer = self._hyper_transformers.get(table_name)
        if hyper_transformer is None:
            hyper_transformer = self._load_hyper_transformer(table_name)
            fields = list(hyper_transformer.transformers.keys())
            hyper_transformer.fit(data[fields])
            self._hyper_transformers[table_name] = hyper_transformer

        hyper_transformer = self._hyper_transformers.get(table_name)
        fields = list(hyper_transformer.transformers.keys())
        return hyper_transformer.transform(data[fields])

    def reverse_transform(self, table_name, data):
        """Reverse the transformed data for a given table.

        Args:
            table_name (str):
                Name of the table to reverse transform.
            data (pandas.DataFrame):
                Data to be reversed.

        Returns:
            pandas.DataFrame
        """
        hyper_transformer = self._hyper_transformers[table_name]
        reversed_data = hyper_transformer.reverse_transform(data)

        for name, dtype in self.get_dtypes(table_name, ids=True).items():
            reversed_data[name] = reversed_data[name].dropna().astype(dtype)

        return reversed_data

    # ################### #
    # Metadata Validation #
    # ################### #

    def _validate_table(self, table_name, table_meta, table_data=None, errors=None):
        """Validate table metadata.

        Validate the type and subtype combination for each field in ``table_meta``.
        If a field has type ``id``, validate that it either is the ``primary_key`` or
        has a ``ref`` entry.

        If the table has ``primary_key``, make sure that the corresponding field exists
        and its type is ``id``.

        If ``table_data`` is provided, also check that the list of columns corresponds
        to the ones indicated in the metadata and that all the dtypes are valid.

        Args:
            table_name (str):
                Name of the table to validate.
            table_meta (dict):
                Metadata of the table to validate.
            table_data (pandas.DataFrame):
                If provided, make sure that the data matches the one described
                on the metadata.

        Raises:
            MetadataError:
                If there is any error in the metadata or the data does not
                match the metadata description.
        """
        errors = [] if errors is None else errors
        dtypes = self.get_dtypes(table_name, ids=True, errors=errors)

        # Primary key field exists and its type is 'id'
        primary_key = table_meta.get('primary_key')
        if primary_key:
            pk_field = table_meta['fields'].get(primary_key)

            if not pk_field:
                errors.append(
                    f'Invalid primary key: "{primary_key}" not found in table "{table_name}"')
            elif pk_field['type'] != 'id':
                errors.append(
                    f'Primary key "{primary_key}" of table "{table_name}" not of type "id"')

        if table_data is not None:
            for column in table_data:
                try:
                    dtype = dtypes.pop(column)
                    table_data[column].dropna().astype(dtype)
                except KeyError:
                    message = 'Unexpected column in table `{}`: `{}`'.format(table_name, column)
                    errors.append(message)
                except ValueError as ve:
                    message = 'Invalid values found in column `{}` of table `{}`: `{}`'.format(
                        column, table_name, ve)
                    errors.append(message)

            # assert all dtypes are in data
            if dtypes:
                errors.append(
                    'Missing columns on table {}: {}.'.format(table_name, list(dtypes.keys()))
                )

    def _validate_circular_relationships(self, parent, children=None, errors=None, parents=None):
        """Validate that there is no circular relatioship in the metadata."""
        errors = [] if errors is None else errors
        parents = set() if parents is None else parents
        if children is None:
            children = self.get_children(parent)

        if parent in children:
            error = 'Circular relationship found for table "{}"'.format(parent)
            errors.append(error)

        for child in children:
            if child in parents:
                break

            parents.add(child)
            self._validate_circular_relationships(
                parent, self.get_children(child), errors, parents)

    def validate(self, tables=None):
        """Validate this metadata.

        Validate the metadata of each table:

            * If ``tables`` are provided or they have been loaded, check
              that all the metadata tables exists in the ``tables`` dictionary.
            * Validate the type/subtype combination for each field and
              if a field of type ``id`` exists it must be the ``primary_key``
              or must have a ``ref`` entry.
            * If ``primary_key`` entry exists, check that it's an existing
              field and its type is ``id``.
            * If ``tables`` are provided or they have been loaded, check
              all the data types for the table correspond to each column and
              all the data types exists on the table.
            * Validate that there is no circular relatioship in the metadata.
            * Check that all the tables have at most one parent.

        Args:
            tables (bool, dict):
                If a dict of table is passed, validate that the columns and
                dtypes match the metadata. If ``True`` is passed, load the
                tables from the Metadata instead. If ``None``, omit the data
                validation. Defaults to ``None``.
        """
        tables_meta = self._metadata.get('tables')
        if not tables_meta:
            raise MetadataError('"tables" entry not found in Metadata.')

        if tables and not isinstance(tables, dict):
            tables = self.load_tables()

        errors = []
        for table_name, table_meta in tables_meta.items():
            if tables:
                table = tables.get(table_name)
                if table is None:
                    errors.append('Table `{}` not found in tables'.format(table_name))

            else:
                table = None

            self._validate_table(table_name, table_meta, table, errors)
            self._validate_circular_relationships(table_name, errors=errors)

        if errors:
            raise MetadataError('Invalid Metadata specification:\n - ' + '\n - '.join(errors))

    def _check_field(self, table, field, exists=False):
        """Validate the existance of the table and existance (or not) of field."""
        table_fields = self.get_fields(table)
        if exists and (field not in table_fields):
            raise ValueError('Field "{}" does not exist in table "{}"'.format(field, table))

        if not exists and (field in table_fields):
            raise ValueError('Field "{}" already exists in table "{}"'.format(field, table))

    # ################# #
    # Metadata Creation #
    # ################# #

    def add_field(self, table, field, field_type, field_subtype=None, properties=None):
        """Add a new field to the indicated table.

        Args:
            table (str):
                Table name to add the new field, it must exist.
            field (str):
                Field name to be added, it must not exist.
            field_type (str):
                Data type of field to be added. Required.
            field_subtype (str):
                Data subtype of field to be added. Optional.
                Defaults to ``None``.
            properties (dict):
                Extra properties of field like: ref, format, min, max, etc. Optional.
                Defaults to ``None``.

        Raises:
            ValueError:
                If the table does not exist or it already contains the field.
        """
        self._check_field(table, field, exists=False)

        field_details = {
            'type': field_type
        }

        if field_subtype:
            field_details['subtype'] = field_subtype

        if properties:
            field_details.update(properties)

        self._metadata['tables'][table]['fields'][field] = field_details

    @staticmethod
    def _get_key_subtype(field_meta):
        """Get the appropriate key subtype."""
        field_type = field_meta['type']

        if field_type == 'categorical':
            field_subtype = 'string'

        elif field_type in ('numerical', 'id'):
            field_subtype = field_meta['subtype']
            if field_subtype not in ('integer', 'string'):
                raise ValueError(
                    'Invalid field "subtype" for key field: "{}"'.format(field_subtype)
                )

        else:
            raise ValueError(
                'Invalid field "type" for key field: "{}"'.format(field_type)
            )

        return field_subtype

    def set_primary_key(self, table, field):
        """Set the primary key field of the indicated table.

        The field must exist and either be an integer or categorical field.

        Args:
            table (str):
                Name of the table where the primary key will be set.
            field (str):
                Name of the field to be used as the new primary key.

        Raises:
            ValueError:
                If the table or the field do not exist or if the field has an
                invalid type or subtype.
        """
        self._check_field(table, field, exists=True)

        field_meta = self.get_fields(table).get(field)
        field_subtype = self._get_key_subtype(field_meta)

        table_meta = self._metadata['tables'][table]
        table_meta['fields'][field] = {
            'type': 'id',
            'subtype': field_subtype
        }
        table_meta['primary_key'] = field

    def add_relationship(self, parent, child, foreign_key=None, validate=True):
        """Add a new relationship between the parent and child tables.

        The relationship is created by adding a reference (``ref``) on the ``foreign_key``
        field of the ``child`` table pointing at the ``parent`` primary key.

        Args:
            parent (str):
                Name of the parent table.
            child (str):
                Name of the child table.
            foreign_key (str):
                Field in the child table through which the relationship is created.
                If ``None``, use the parent primary key name.
            validate (bool):
                Whether to validate metadata after adding this relationship or not.
                Defaults to ``True``.

        Raises:
            ValueError:
                If any of the following happens:
                    * The parent table does not exist.
                    * The child table does not exist.
                    * The parent table does not have a primary key.
                    * The foreign_key field already exists in the child table.
                    * The child table already has a parent.
                    * The new relationship closes a relationship circle.
        """
        # Make a backup
        metadata_backup = copy.deepcopy(self._metadata)

        # Validate tables exists
        self.get_table_meta(parent)
        self.get_table_meta(child)

        # Validate field names
        primary_key = self.get_primary_key(parent)
        if not primary_key:
            raise ValueError('Parent table "{}" does not have a primary key'.format(parent))

        if foreign_key is None:
            foreign_key = primary_key

        parent_key_meta = copy.deepcopy(self.get_field_meta(parent, primary_key))
        child_key_meta = copy.deepcopy(self.get_field_meta(child, foreign_key))

        # Validate relationships
        child_ref = child_key_meta.get('ref')
        if child_ref:
            raise ValueError(
                'Field "{}.{}" already defines a relationship'.format(child, foreign_key))

        # Make sure that the parent key is an id
        if parent_key_meta['type'] != 'id':
            parent_key_meta['subtype'] = self._get_key_subtype(parent_key_meta)
            parent_key_meta['type'] = 'id'

        # Update the child key meta
        child_key_meta['subtype'] = self._get_key_subtype(child_key_meta)
        child_key_meta['type'] = 'id'
        child_key_meta['ref'] = {
            'table': parent,
            'field': primary_key
        }

        # Make sure that key subtypes are the same
        if child_key_meta['subtype'] != parent_key_meta['subtype']:
            raise ValueError('Parent and Child key subtypes mismatch')

        self._metadata['tables'][parent]['fields'][primary_key] = parent_key_meta
        self._metadata['tables'][child]['fields'][foreign_key] = child_key_meta

        # Re-analyze the relationships
        self._analyze_relationships()

        if validate:
            try:
                self.validate()
            except MetadataError:
                self._metadata = metadata_backup
                raise

    def _get_field_details(self, data, fields):
        """Get or build all the fields metadata.

        Analyze a ``pandas.DataFrame`` to build a ``dict`` with the name of the column, and
        their data type and subtype. If ``columns`` are provided, only those columns will be
        analyzed.

        Args:
            data (pandas.DataFrame):
                Table to be analyzed.
            fields (set):
                Set of field names or field specifications.

        Returns:
            dict:
                Dict of valid fields.

        Raises:
            TypeError:
                If a field specification is not a str or a dict.
            ValueError:
                If a column from the data analyzed is an unsupported data type or
        """
        fields_metadata = dict()
        for field in fields:
            dtype = data[field].dtype
            field_template = self._FIELD_TEMPLATES.get(dtype.kind)
            if not field_template:
                raise ValueError('Unsupported dtype {} in column {}'.format(dtype, field))

            field_details = copy.deepcopy(field_template)
            fields_metadata[field] = field_details

        return fields_metadata

    def add_table(self, name, data=None, fields=None, fields_metadata=None,
                  primary_key=None, parent=None, foreign_key=None, constraints=None):
        """Add a new table to this metadata.

        ``fields`` list can be a mixture of field names, which will be build automatically
        from the data, or dictionaries specifying the field details. If a field needs to be
        analyzed, data has to be also passed.

        If ``parent`` is given, a relationship will be established between this table
        and the specified parent.

        Args:
            name (str):
                Name of the new table.
            data (str or pandas.DataFrame):
                Table to be analyzed or path to the csv file.
                If it's a relative path, use ``root_path`` to find the file.
                Only used if fields is not ``None``.
                Defaults to ``None``.
            fields (list):
                List of field names to build. If ``None`` is given, all the fields
                found in the data will be used.
                Defaults to ``None``.
            fields_metadata (dict):
                Metadata to be used when creating fields. This will overwrite the
                metadata built from the fields found in data.
                Defaults to ``None``.
            primary_key (str):
                Field name to add as primary key, it must not exists. Defaults to ``None``.
            parent (str):
                Table name to refere a foreign key field. Defaults to ``None``.
            foreign_key (str):
                Foreign key field name to ``parent`` table primary key. Defaults to ``None``.
            constraints (list[Constraint, dict]):
                List of Constraint objects or dicts representing the constraints for the
                given table.

        Raises:
            ValueError:
                If the table ``name`` already exists or ``data`` is not passed and
                fields need to be built from it.
        """
        if name in self.get_tables():
            raise ValueError('Table "{}" already exists.'.format(name))

        path = None
        if data is not None:
            if isinstance(data, str):
                path = data
                if not os.path.isabs(data):
                    data = os.path.join(self.root_path, data)

                data = pd.read_csv(data)

            fields = fields or data.columns
            if fields_metadata:
                fields = [field for field in fields if field not in fields_metadata]
            else:
                fields_metadata = dict()

            fields_metadata.update(self._get_field_details(data, fields))

        elif fields_metadata is None:
            fields_metadata = dict()

        table_metadata = {'fields': fields_metadata}
        if path:
            table_metadata['path'] = path

        self._metadata['tables'][name] = table_metadata

        if constraints:
            meta_constraints = []
            for constraint in constraints:
                if isinstance(constraint, Constraint):
                    meta_constraints.append(constraint.to_dict())
                else:
                    meta_constraints.append(constraint)

            table_metadata['constraints'] = meta_constraints

        try:
            if primary_key:
                self.set_primary_key(name, primary_key)

            if parent:
                self.add_relationship(parent, name, foreign_key)

        except ValueError:
            # Cleanup
            del self._metadata['tables'][name]
            raise

    # ###################### #
    # Metadata Serialization #
    # ###################### #

    def to_dict(self):
        """Get a dict representation of this metadata.

        Returns:
            dict:
                dict representation of this metadata.
        """
        return copy.deepcopy(self._metadata)

    def to_json(self, path):
        """Dump this metadata into a JSON file.

        Args:
            path (str):
                Path of the JSON file where this metadata will be stored.
        """
        with open(path, 'w') as out_file:
            json.dump(self._metadata, out_file, indent=4)

    def __repr__(self):
        tables = self.get_tables()
        relationships = [
            '    {}.{} -> {}.{}'.format(
                table, foreign_key,
                parent, self.get_primary_key(parent)
            )
            for table in tables
            for parent in list(self.get_parents(table))
            for foreign_key in self.get_foreign_keys(parent, table)
        ]

        return (
            "Metadata\n"
            "  root_path: {}\n"
            "  tables: {}\n"
            "  relationships:\n"
            "{}"
        ).format(
            self.root_path,
            tables,
            '\n'.join(relationships)
        )

    def visualize(self, path=None, names=True, details=True):
        """Plot metadata usign graphviz.

        Generate a plot using graphviz.
        If a ``path`` is provided save the output into a file.

        Args:
            path (str):
                Output file path to save the plot. It requires a graphviz
                supported extension. If ``None`` do not save the plot and
                just return the ``graphviz.Digraph`` object.
                Defaults to ``None``.
            names (bool):
                Whether to add names to the diagram or not. Defaults to ``True``
        """
        return visualization.visualize(self, path, names=names, details=details)



"""Functions to profile performance of rdt2 Transformers."""

# pylint: disable=W0212

import multiprocessing as mp
import timeit
import tracemalloc
from copy import deepcopy

import pandas as pd


def _profile_time(transformer, method_name, dataset, iterations=10, copy=False):
    total_time = 0
    for _ in range(iterations):
        if copy:
            transformer_copy = deepcopy(transformer)
            method = getattr(transformer_copy, method_name)

        else:
            method = getattr(transformer, method_name)

        start_time = timeit.default_timer()
        method(dataset)
        total_time += timeit.default_timer() - start_time

    return total_time / iterations


def _set_memory_for_method(method, dataset, peak_memory):
    tracemalloc.start()
    method(dataset)
    peak_memory.value = tracemalloc.get_traced_memory()[1]
    tracemalloc.stop()
    tracemalloc.clear_traces()


def _profile_memory(method, dataset):
    ctx = mp.get_context('spawn')
    peak_memory = ctx.Value('i', 0)
    profiling_process = ctx.Process(
        target=_set_memory_for_method,
        args=(method, dataset, peak_memory)
    )
    profiling_process.start()
    profiling_process.join()
    return peak_memory.value


def profile_transformer(transformer, dataset_generator, transform_size, fit_size=None):
    """Profile a Transformer on a dataset.

    This function will get the total time and peak memory
    for the ``fit``, ``transform`` and ``reverse_transform``
    methods of the provided transformer against the provided
    dataset.

    Args:
        transformer (Transformer):
            Transformer instance.
        dataset_generator (DatasetGenerator):
            DatasetGenerator instance.
        transform_size (int):
            Number of rows to generate for ``transform`` and ``reverse_transform``.
        fit_size (int or None):
            Number of rows to generate for ``fit``. If None, use ``transform_size``.

    Returns:
        pandas.Series:
            Series containing the time and memory taken by ``fit``, ``transform``,
            and ``reverse_transform`` for the transformer.
    """
    fit_size = fit_size or transform_size
    fit_dataset = pd.Series(dataset_generator.generate(fit_size))
    replace = transform_size > fit_size
    transform_dataset = fit_dataset.sample(transform_size, replace=replace)

    try:
        fit_time = _profile_time(transformer, 'fit', fit_dataset, copy=True)
        fit_memory = _profile_memory(transformer.fit, fit_dataset)
        transformer.fit(fit_dataset)

        transform_time = _profile_time(transformer, 'transform', transform_dataset)
        transform_memory = _profile_memory(transformer.transform, transform_dataset)

        reverse_dataset = transformer.transform(transform_dataset)
        reverse_time = _profile_time(transformer, 'reverse_transform', reverse_dataset)
        reverse_memory = _profile_memory(transformer.reverse_transform, reverse_dataset)
    except TypeError:
        # temporarily support both old and new style transformers
        fit_time = _profile_time(transformer, '_fit', fit_dataset, copy=True)
        fit_memory = _profile_memory(transformer._fit, fit_dataset)
        transformer._fit(fit_dataset)

        transform_time = _profile_time(transformer, '_transform', transform_dataset)
        transform_memory = _profile_memory(transformer._transform, transform_dataset)

        reverse_dataset = transformer._transform(transform_dataset)
        reverse_time = _profile_time(transformer, '_reverse_transform', reverse_dataset)
        reverse_memory = _profile_memory(transformer._reverse_transform, reverse_dataset)

    return pd.Series({
        'Fit Time': fit_time,
        'Fit Memory': fit_memory,
        'Transform Time': transform_time,
        'Transform Memory': transform_memory,
        'Reverse Transform Time': reverse_time,
        'Reverse Transform Memory': reverse_memory
    })



import os
from pathlib import Path

import numpy as np
from matplotlib import rc
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

from syntherela.visualisations.utils import get_x_tick_width_coef

rc("font", **{"family": "serif", "serif": ["Times"]})
rc("text", usetex=True)


def visualize_parent_child_multi_table(all_results, datasets, methods, **kwargs):
    for dataset in datasets:
        metrics = kwargs.get(
            "detection_metrics",
            [
                metric
                for metric in list(
                    all_results[dataset][methods[0]]["multi_table_metrics"].keys()
                )
                if "ParentChildDetection" in metric
            ],
        )
        metric_names = kwargs.get("detection_metric_names", metrics)

        aggregation_metrics = kwargs.get(
            "aggregation_metrics",
            [
                metric
                for metric in list(
                    all_results[datasets[0]][methods[0]]["multi_table_metrics"].keys()
                )
                if "ParentChildAggregationDetection" in metric
            ],
        )

        save_figs = kwargs.get("save_figs", False)
        save_figs_path = kwargs.get("save_figs_path", "./figs")
        save_figs_path = Path(save_figs_path) / "multi_table" / "detection"

        method_order = kwargs.get(
            "method_order", ["SDV", "RCTGAN", "MOSTLYAI", "REALTABFORMER"]
        )
        if method_order is not None:
            methods = [method for method in method_order if method in methods]
            methods += sorted(
                [method for method in methods if method not in method_order]
            )

        if len(methods) == 0 or len(metrics) == 0:
            continue
        tables = all_results[dataset][methods[0]]["multi_table_metrics"][
            metrics[0]
        ].keys()
        for table in tables:
            agg_metrics = []
            if (
                aggregation_metrics
                and table
                in all_results[dataset][methods[0]]["multi_table_metrics"][
                    aggregation_metrics[0]
                ]
            ):
                agg_metrics = aggregation_metrics

            N = len(metrics + agg_metrics)  # number of metrics
            M = len(methods)  # number of methods
            ind = np.arange(M)  # the x locations for the groups
            width = 0.15  # the width of the bars

            fig, ax = plt.subplots(figsize=(10, 7))
            # set dpi
            fig.dpi = 300

            colors = plt.cm.viridis(np.linspace(0, 1, N))  # create a color map

            min_mean = 1
            for j, metric in enumerate(metrics):
                metric_means = [
                    all_results[dataset][method]["multi_table_metrics"][metric][table][
                        "accuracy"
                    ]
                    for method in methods
                ]
                min_mean = min(min_mean, min(metric_means))
                metric_ses = [
                    all_results[dataset][method]["multi_table_metrics"][metric][table][
                        "SE"
                    ]
                    for method in methods
                ]
                # baseline_means = np.array([all_results[dataset][method]['multi_table_metrics'][metric][table]["baseline_mean"] for method in methods])
                # baseline_ses = np.array([all_results[dataset][method]['multi_table_metrics'][metric][table]["baseline_se"] for method in methods])

                baseline_means = np.array([0.5 for method in methods])
                baseline_ses = np.array([0.00 for method in methods])

                ax.bar(
                    ind + width * j,
                    metric_means,
                    width,
                    yerr=metric_ses,
                    color=colors[j],
                )
                ax.hlines(
                    baseline_means,
                    ind + width * j - width / 2,
                    ind + width * j + width / 2,
                    color="k",
                )  # , linestyle='--')
                ax.hlines(
                    baseline_means + baseline_ses,
                    ind + width * j - width / 2,
                    ind + width * j + width / 2,
                    color="k",
                    linestyle="--",
                )
                ax.hlines(
                    baseline_means - baseline_ses,
                    ind + width * j - width / 2,
                    ind + width * j + width / 2,
                    color="k",
                    linestyle="--",
                )

            for j, agg_metric in enumerate(agg_metrics):
                metric_means = [
                    all_results[dataset][method]["multi_table_metrics"][agg_metric][
                        table
                    ]["accuracy"]
                    for method in methods
                ]
                min_mean = min(min_mean, min(metric_means))
                metric_ses = [
                    all_results[dataset][method]["multi_table_metrics"][agg_metric][
                        table
                    ]["SE"]
                    for method in methods
                ]
                # baseline_means = np.array([all_results[dataset][method]['multi_table_metrics'][agg_metric][table]["baseline_mean"] for method in methods])
                # baseline_ses = np.array([all_results[dataset][method]['multi_table_metrics'][agg_metric][table]["baseline_se"] for method in methods])

                baseline_means = np.array([0.5 for method in methods])
                baseline_ses = np.array([0.00 for method in methods])

                ax.bar(
                    ind + width * (j + len(metrics)),
                    metric_means,
                    width,
                    yerr=metric_ses,
                    color=colors[j + len(metrics)],
                )
                ax.hlines(
                    baseline_means,
                    ind + width * (j + len(metrics)) - width / 2,
                    ind + width * (j + len(metrics)) + width / 2,
                    color="k",
                )
                ax.hlines(
                    baseline_means + baseline_ses,
                    ind + width * (j + len(metrics)) - width / 2,
                    ind + width * (j + len(metrics)) + width / 2,
                    color="k",
                    linestyle="--",
                )
                ax.hlines(
                    baseline_means - baseline_ses,
                    ind + width * (j + len(metrics)) - width / 2,
                    ind + width * (j + len(metrics)) + width / 2,
                    color="k",
                    linestyle="--",
                )

            ax.set_ylabel("Means")
            x_tick_width_coef = get_x_tick_width_coef(N)
            ax.set_xticks(ind + x_tick_width_coef * width)
            rotation = 20 if len(methods) > 6 else 0
            ax.set_xticklabels(methods, fontsize=10, rotation=rotation)

            y_min = 0.4 if min_mean > 0.4 else np.floor((min_mean - 0.1) * 10) / 10
            ax.set_ylim(y_min, 1.3)
            ax.set_yticks(np.arange(y_min, 1.01, 0.1))
            ax.set_ylabel("Classification Accuracy")

            # Create a legend
            custom_lines = [Line2D([0], [0], color=colors[i], lw=4) for i in range(N)]
            ax.legend(
                custom_lines, metric_names + agg_metrics, loc="upper left"
            )  # move the legend

            ax.axhline(y=0.5, color="red", linestyle="--", linewidth=1)

            # set title
            plt.title(f"Dataset {dataset}, table {table}")

            if save_figs:
                os.makedirs(save_figs_path, exist_ok=True)
                plt.savefig(
                    f"{save_figs_path}/{dataset}_{table}_parent_child_detection.png",
                    dpi=300,
                )


def visualize_multi_table(all_results, datasets, methods, **kwargs):
    save_figs = kwargs.get("save_figs", False)
    save_figs_path = kwargs.get("save_figs_path", "./figs/")
    save_figs_path = Path(save_figs_path) / "multi_table" / "detection"

    methods_all = methods.copy()

    for dataset in datasets:
        methods = methods_all.copy()
        method_order = kwargs.get(
            "method_order",
            [
                "SDV",
                "RCTGAN",
                "REALTABFORMER",
                "MOSTLYAI",
                "GRETEL_ACTGAN",
                "GRETEL_LSTM",
            ],
        )
        metrics = kwargs.get(
            "detection_metrics",
            [
                metric
                for metric in list(
                    all_results[dataset][method_order[1]]["multi_table_metrics"].keys()
                )
                if "singletable" not in metric.lower()
                and "parent" not in metric.lower()
                and "detection" in metric.lower()
            ],
        )
        metric_names = kwargs.get("detection_metric_names", metrics)

        if method_order is not None:
            methods = [
                method
                for method in method_order
                if method in methods and method in all_results[dataset]
            ]
            methods += sorted(
                [method for method in methods if method not in method_order]
            )
        if len(methods) == 0 or len(metrics) == 0:
            continue

        N = len(metrics)  # number of metrics
        M = len(methods)  # number of methods
        ind = np.arange(M)  # the x locations for the groups
        width = 0.15  # the width of the bars

        fig, ax = plt.subplots(figsize=(10, 7))
        # set dpi
        fig.dpi = 300

        colors = plt.cm.viridis(np.linspace(0, 1, N))  # create a color map

        min_mean = 1

        for j, metric in enumerate(metrics):
            metric_means = [
                all_results[dataset][method]["multi_table_metrics"][metric]["accuracy"]
                for method in methods
            ]
            min_mean = min(min_mean, min(metric_means))
            metric_ses = [
                all_results[dataset][method]["multi_table_metrics"][metric]["SE"]
                for method in methods
            ]

            baseline_means = np.array([0.5 for method in methods])
            baseline_ses = np.array([0.00 for method in methods])

            ax.bar(
                ind + width * j, metric_means, width, yerr=metric_ses, color=colors[j]
            )
            # draw a horizontal line for the baseline and standard error
            ax.hlines(
                baseline_means,
                ind + width * j - width / 2,
                ind + width * j + width / 2,
                color="k",
            )  # , linestyle='--')
            ax.hlines(
                baseline_means + baseline_ses,
                ind + width * j - width / 2,
                ind + width * j + width / 2,
                color="k",
                linestyle="--",
            )
            ax.hlines(
                baseline_means - baseline_ses,
                ind + width * j - width / 2,
                ind + width * j + width / 2,
                color="k",
                linestyle="--",
            )

        ax.set_ylabel("Means")
        x_tick_width_coef = get_x_tick_width_coef(N)
        ax.set_xticks(ind + x_tick_width_coef * width)
        rotation = 20 if len(methods) > 6 else 0
        ax.set_xticklabels(methods, fontsize=10, rotation=rotation)

        y_min = 0.4 if min_mean > 0.4 else np.floor((min_mean - 0.1) * 10) / 10
        ax.set_ylim(y_min, 1.3)
        ax.set_yticks(np.arange(y_min, 1.01, 0.1))
        ax.set_ylabel("Classification Accuracy")

        # Create a legend
        custom_lines = [Line2D([0], [0], color=colors[i], lw=4) for i in range(N)]
        ax.legend(custom_lines, metric_names, loc="upper left")  # move the legend

        ax.axhline(y=0.5, color="red", linestyle="--", linewidth=1)

        # set title
        plt.title(f"Dataset {dataset}")

        if save_figs:
            os.makedirs(save_figs_path, exist_ok=True)
            plt.savefig(save_figs_path / f"{dataset}_multi_table_detection.png")



import os
import json
import pickle
import argparse
from pathlib import Path

import pandas as pd
from syntherela.metadata import Metadata
from syntherela.data import save_tables, load_tables, remove_sdv_columns

from ClavaDDPM.preprocess_utils import table_label_decode, reconstruct_dates


def revert_ids(df, metadata, table_name):
    primary_key = metadata.get_primary_key(table_name)
    if primary_key is not None:
        df[primary_key] = df[f"{table_name}_id"]
        df.drop(columns=[f"{table_name}_id"], inplace=True)
    else:
        df.drop(columns=[f"{table_name}_id"], inplace=True)

    for parent in metadata.get_parents(table_name):
        foreign_keys = metadata.get_foreign_keys(parent, table_name)
        assert len(foreign_keys) == 1
        foreign_key = foreign_keys[0]
        df[foreign_key] = df[f"{parent}_id"]
        df.drop(columns=[f"{parent}_id"], inplace=True)

    return df


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("dataset_name", type=str, help="Name of the dataset")
    parser.add_argument("--run-id", type=int, default=1, help="Run id")
    parser.add_argument(
        "--real-data-path",
        type=str,
        default="data/original/",
        help="Path to the original data.",
    )
    parser.add_argument(
        "--synthetic-data-path",
        type=str,
        default="data/synthetic/",
        help="Path to the original data.",
    )

    return parser.parse_args()


def main(args):
    run = str(args.run_id)
    dataset_name = args.dataset_name
    real_data_path = args.real_data_path
    synthetic_data_path = args.synthetic_data_path

    model_name = "CLAVADDPM"

    metadata = Metadata().load_from_json(
        Path(real_data_path) / f"{dataset_name}/metadata.json"
    )

    tables = load_tables(Path(real_data_path) / f"{dataset_name}", metadata)

    tables, metadata = remove_sdv_columns(tables, metadata)

    processed_data_path = os.path.join("ClavaDDPM", "complex_data", dataset_name)
    generated_data_path = os.path.join(
        "ClavaDDPM", f"clavaDDPM_workspace_run{run}", dataset_name
    )
    synthetic_data_path = os.path.join(synthetic_data_path, dataset_name)

    if os.path.exists(os.path.join(processed_data_path, "first_dates.json")):
        with open(os.path.join(processed_data_path, "first_dates.json"), "r") as f:
            first_dates = json.load(f)

    tables_synthetic = dict()
    for table_name in metadata.get_tables():
        table_meta = metadata.get_table_meta(table_name)["columns"]
        table_path = os.path.join(
            generated_data_path, table_name, "_final", f"{table_name}_synthetic.csv"
        )
        df = pd.read_csv(table_path)

        datetime_columns = metadata.get_column_names(table_name, sdtype="datetime")
        numerical_columns = metadata.get_column_names(table_name, sdtype="numerical")

        le_path = os.path.join(processed_data_path, f"{table_name}_label_encoders.pkl")
        with open(le_path, "rb") as f:
            label_encoders = pickle.load(f)

        for column in label_encoders.keys():
            if column in df.columns:
                df[column] = df[column].astype(int)

        df = table_label_decode(df, label_encoders)

        if len(datetime_columns):
            for col in datetime_columns:
                first_date_str = first_dates[table_name][col]
                df[col] = reconstruct_dates(df[col], first_date_str)
                df[col] = pd.to_datetime(df[col], format="%y%m%d")
                column_format = table_meta[col].get("datetime_format", "%Y-%m-%d")
                df[col] = df[col].dt.strftime(column_format)

        for col in numerical_columns:
            dtype = table_meta[col]["computer_representation"]
            if dtype == "Int64":
                df[col] = df[col].round().astype(dtype)

        df = revert_ids(df, metadata, table_name)
        tables_synthetic[table_name] = df

    save_data_path = os.path.join(synthetic_data_path, model_name, run, "sample1")
    save_tables(tables_synthetic, save_data_path)


if __name__ == "__main__":
    args = parse_args()
    main(args)



from typing import Tuple

import numpy as np
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import mean_squared_error, roc_auc_score
from sdmetrics.base import BaseMetric

from syntherela.metadata import drop_ids
from syntherela.utils import CustomHyperTransformer


class MachineLearningEfficacyMetric(BaseMetric):
    def __init__(
        self,
        target: Tuple[str, str],
        classifier_cls,
        classifier_args={},
        random_state=None,
        feature_engineering_function=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.target = target
        self.classifier_cls = classifier_cls
        self.classifier_args = classifier_args
        self.random_state = random_state
        self.name = f"{type(self).__name__}-{classifier_cls.__name__}"
        self.feature_engineering_function = feature_engineering_function

    def prepare_data(self, X, ht=None, **kwargs):
        if ht is None:
            ht = CustomHyperTransformer()
            ht.fit(X.copy())
        transformed_data = ht.transform(X.copy())
        # replace infinite values with NaN
        transformed_data.replace([np.inf, -np.inf], np.nan, inplace=True)
        return transformed_data, ht

    def score(self, model, X, y):
        # if classifier is a regressor, compute RMSE, else AUCROC
        if hasattr(model, "predict_proba"):
            probs = model.predict_proba(X)
            if probs.shape[1] > 2:
                # calculate AUCROC
                return roc_auc_score(y, probs, multi_class="ovr", average="macro")
            else:
                return roc_auc_score(y, probs[:, 1])
        else:
            # calculate RMSE
            y_pred = model.predict(X)
            return -np.sqrt(mean_squared_error(y, y_pred))

    def compute(self, X_train, y_train, X_test, y_test, m=100, **kwargs):
        np.random.seed(self.random_state)
        model_full_train_set = Pipeline(
            [
                ("imputer", SimpleImputer()),
                ("scaler", StandardScaler()),
                ("clf", self.classifier_cls(**self.classifier_args)),
            ]
        )
        model_full_train_set.fit(X_train, y_train)
        score_full_train_set = self.score(model_full_train_set, X_test, y_test)

        # Bootstrap to estimate the standard error
        scores = []
        models = []
        for bootstrap_idx in range(m):
            np.random.seed(self.random_state + bootstrap_idx)
            indices = np.random.choice(len(X_train), len(X_train), replace=m > 1)
            model = Pipeline(
                [
                    ("imputer", SimpleImputer()),
                    ("scaler", StandardScaler()),
                    ("clf", self.classifier_cls(**self.classifier_args)),
                ]
            )

            X_train_boot = X_train.iloc[indices]
            y_train_boot = y_train.iloc[indices]

            model.fit(X_train_boot, y_train_boot)
            score = self.score(model, X_test, y_test)
            scores.append(score)
            models.append(model)

        return (
            model_full_train_set,
            models,
            scores,
            score_full_train_set,
            np.std(scores) / np.sqrt(m),
        )

    def get_target_table(self, data, target, metadata):
        target_table, target_column, _ = target
        X, y = (
            data[target_table].drop(columns=target_column),
            data[target_table][target_column],
        )
        X = drop_ids(X, metadata.to_dict()["tables"][target_table])
        return X, y

    def run(
        self,
        real_data,
        synthetic_data,
        metadata,
        test_data,
        m=100,
        feature_importance=True,
        score_real=None,
        se_real=None,
        feature_importance_real=None,
        **kwargs,
    ):
        if self.feature_engineering_function:
            X_real, y_real = self.feature_engineering_function(real_data, metadata)
            X_synthetic, y_synthetic = self.feature_engineering_function(
                synthetic_data, metadata
            )
            X_test, y_test = self.feature_engineering_function(test_data, metadata)
        else:
            X_real, y_real = self.get_target_table(real_data, self.target, metadata)
            X_synthetic, y_synthetic = self.get_target_table(
                synthetic_data, self.target, metadata
            )
            X_test, y_test = self.get_target_table(test_data, self.target, metadata)
        X_real = X_real[X_test.columns]
        X_synthetic = X_synthetic[X_test.columns]

        X_real, ht = self.prepare_data(X_real)
        X_synthetic, _ = self.prepare_data(X_synthetic, ht=ht)
        X_test, _ = self.prepare_data(X_test, ht=ht)

        # save the data for feature importance methods
        self.X_real = X_real
        self.y_real = y_real
        self.X_synthetic = X_synthetic
        self.y_synthetic = y_synthetic
        self.X_test = X_test
        self.y_test = y_test

        compute_real = score_real is None or se_real is None

        if compute_real:
            model_real, _, _, score_real, se_real = self.compute(
                X_real, y_real, X_test, y_test, m=m
            )
        (
            model_synthetic,
            models_synthetic,
            score_array_synthetic,
            score_synthetic,
            se_synthetic,
        ) = self.compute(X_synthetic, y_synthetic, X_test, y_test, m=m)
        # compute the baseline score
        difference = score_synthetic - score_real
        importances_syn = []
        if feature_importance:
            if compute_real:
                full_feature_importance_real, _ = self.feature_importance(model_real)
            else:
                full_feature_importance_real = feature_importance_real
            full_feature_importance_syn, feature_names = self.feature_importance(
                model_synthetic
            )
            for model_synthetic in models_synthetic:
                importance_syn, feature_names = self.feature_importance(model_synthetic)
                importances_syn.append(importance_syn)
        else:
            feature_names = []
            full_feature_importance_real = []
            full_feature_importance_syn = []

        return {
            "real_score": score_real,
            "real_score_se": se_real,
            "synthetic_score": score_synthetic,
            "synthetic_score_se": se_synthetic,
            "synthetic_score_array": score_array_synthetic,
            "difference": difference,
            "importance_synthetic": importances_syn,
            "feature_names": feature_names,
            "full_feature_importance_real": full_feature_importance_real,
            "full_feature_importance_synthetic": full_feature_importance_syn,
        }

    def feature_importance(self, model):
        if hasattr(model["clf"], "feature_importances_"):
            importance = model["clf"].feature_importances_
        elif hasattr(model["clf"], "coef_"):
            importance = model["clf"].coef_
        else:
            raise NotImplementedError(
                f"Feature importance not supported for {type(model['clf'])}"
            )

        return importance, self.X_real.columns.tolist()



import os
import argparse

import psycopg2
from tqdm import tqdm
from dotenv import load_dotenv

from relsyndgb.metadata import Metadata
from relsyndgb.data import load_tables, remove_sdv_columns


def legalize_table_name(table_name: str):
    return f'"{table_name}"'

# TABLE GENERATION FUNCTIONS
def get_foreign_key_reference(table_name, field, metadata):
    for relationship in metadata.relationships:
        if relationship['child_table_name'] == table_name and relationship['child_foreign_key'] == field:
            return relationship['parent_table_name'], relationship['parent_primary_key']
    return None, None

def create_table_query(table_name, metadata, varchar_length=255, pk_length=20):
    print("Creating table", table_name, "as", legalize_table_name(table_name))
    fields = metadata.tables[table_name].to_dict()['columns']
    fields_str = ''
    for field, values in fields.items():
        if values['sdtype'] == 'id':
            fields_str += f'"{field}" VARCHAR({pk_length})'
            # check if its the primary key
            table_meta = metadata.to_dict()['tables'][table_name]
            if 'primary_key' in table_meta and table_meta['primary_key'] == field:
                fields_str += ' PRIMARY KEY, '
            else: 
                parent, parent_field = get_foreign_key_reference(table_name, field, metadata)
                fields_str += f' REFERENCES {legalize_table_name(parent)}("{parent_field}"), '
            
        elif values['sdtype'] == 'categorical':
            fields_str += f'"{field}" VARCHAR({varchar_length}), '
        elif values['sdtype'] == 'numerical':
            fields_str += f'"{field}" FLOAT, '
        elif values['sdtype'] == 'boolean':
            fields_str += f'"{field}" BOOLEAN, '
        elif values['sdtype'] == 'datetime':
            fields_str += f'"{field}" DATE, '
        else:
            raise ValueError(f'Unknown type {values["type"]} for field {field}')
    fields_str = fields_str[:-2]
    query = f"CREATE TABLE IF NOT EXISTS {legalize_table_name(table_name)} ({fields_str});"
    return query


def execute_query(query, cursor, connection):
    try:
        # Execute the query
        cursor.execute(query)

        # Commit the transaction
        connection.commit()

    except psycopg2.Error as error:
        # Handle any errors that occur during query execution
        print("Error executing query:", error)


def execute_write_query(query, cursor, connection, values):
        try:
            # Execute the INSERT query with %s placeholder
            cursor.execute(query, values)
    
            # Commit the transaction
            connection.commit()
    
            #print("Query executed successfully")
    
        except psycopg2.Error as error:
            raise error
            # Handle any errors that occur during query execution
            # print("Error executing query:", error)


# psycopg2.extensions.register_adapter(float, lambda x: 'NULL' if pd.isna(x) else float(x))        
def insert_batch_rows(table_name, df, batch_size=100):
    insert_query = f"INSERT INTO {legalize_table_name(table_name)} VALUES (" + "%s,"*(len(df.columns)-1) + "%s)"
    rows = [list(row) for _, row in df.iterrows()]
    for row in rows:
        for i, value in enumerate(row):
            if str(value) == 'NaT':
                row[i] = None
    # Insert rows in batches
    total_rows = len(rows)
    batch_size = total_rows if batch_size > total_rows else batch_size
    for i in tqdm(range(0, total_rows, batch_size)):
        batch = rows[i:i+batch_size] if i+batch_size <= total_rows else rows[i:]
        cursor.executemany(insert_query, batch)
        connection.commit()
        # break


if __name__ == "__main__":
    load_dotenv()

    args = argparse.ArgumentParser()
    args.add_argument("--dataset-name", type=str, required=True)
    args.add_argument("--varchar-length", type=int, default=255)
    args.add_argument("--pk-length", type=int, default=40)
    args.add_argument("--drop", default=True, action='store_true')
    args.add_argument("--partition", type=int, default=-1)
    args = args.parse_args()

    # Connect to the database
    print("Connecting to the PostgreSQL database...")
    connection = psycopg2.connect(host=os.environ.get('PG_HOST'),
                            port=os.environ.get('PG_PORT'),
                            user=os.environ.get('PG_USER'),
                            password=os.environ.get('PG_PASSWORD'),
                            dbname=os.environ.get('DBNAME'),
                            # sslmode='require',
                            )

    cursor = connection.cursor()
    print("Connection successful")

    dataset_name = args.dataset_name
    print("Creating tables for", dataset_name)
    # CREATE TABLES
    # clear the database
    if args.drop:
        execute_query("DROP SCHEMA public CASCADE; CREATE SCHEMA public;", cursor, connection)

    # create the tables
    metadata = Metadata().load_from_json(f'/Users/martinjurkovic/Documents/github_projects/relsyndgb/data/downloads/{dataset_name}/metadata.json')
    tables = load_tables(f'/Users/martinjurkovic/Documents/github_projects/relsyndgb/data/downloads/{dataset_name}', metadata)
    tables, metadata = remove_sdv_columns(tables, metadata)
    table_names = list()
    # add parent tables first
    for relationship in metadata.relationships:
        table_names.append(relationship['parent_table_name'])
        
    for relationship in metadata.relationships:
        table_names.append(relationship['child_table_name'])

    table_names = list((dict.fromkeys(table_names)))
    for table_name in table_names:

        if args.partition <= 0:
            # drop the table if it exists
            query = f"DROP TABLE IF EXISTS {legalize_table_name(table_name)} CASCADE;"
            execute_query(query, cursor, connection)
        # create the table
        query = create_table_query(table_name, metadata, 
                                varchar_length=args.varchar_length, 
                                pk_length=args.pk_length)
        execute_query(query, cursor, connection)
        # insert the rows
        fields = metadata.to_dict()['tables'][table_name]['columns']
        # reorder the columns of tables to be like in the metadata
        table_data = tables[table_name][list(fields.keys())]
        # select the partition
        if args.partition > -1:
            p = args.partition
            table_data = table_data[100000 * p:100000 * (p+1)]
        # insert rows in batches
        insert_batch_rows(table_name, table_data, batch_size=10000)



import os
import math
from pathlib import Path

import numpy as np
from matplotlib import rc
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D

from syntherela.visualisations.utils import (
    get_x_tick_width_coef,
    prettify_dataset_name,
    prettify_method_name,
    get_color,
    get_dataset_info,
)

rc("font", **{"family": "serif", "serif": ["Times"]})
rc("text", usetex=True)


def visualize_single_table_distance_metrics(
    granularity_level,
    metric_type,
    all_results,
    datasets,
    methods,
    title=True,
    log_scale=False,
    fontsize=35,
    **kwargs,
):
    for dataset in datasets:
        base_metrics, base_metric_names, save_figs, save_figs_path, methods = (
            get_dataset_info(
                granularity_level, metric_type, all_results, dataset, methods, **kwargs
            )
        )

        for base_metric, base_metric_name in zip(base_metrics, base_metric_names):
            if len(methods) == 0:
                continue

            tables = all_results[dataset][list(all_results[dataset].keys())[0]][
                "single_table_metrics"
            ][base_metric].keys()

            N = len(methods)  # number of methods
            M = len(tables)  # number of tables

            ind = np.arange(M)
            width = 0.12

            fig, ax = plt.subplots(figsize=(10, 7))
            # set dpi
            fig.dpi = 300

            if log_scale:
                ax.set_yscale("log")

            colors = [get_color(method) for method in methods]
            if None in colors:
                colors = plt.cm.viridis(np.linspace(0.5, 1, N))

            max_value = 0

            for j, method in enumerate(methods):
                method_means = [
                    all_results[dataset][method]["single_table_metrics"][base_metric][
                        table
                    ]["value"]
                    for table in tables
                ]
                method_ses = [
                    all_results[dataset][method]["single_table_metrics"][base_metric][
                        table
                    ]["bootstrap_se"]
                    for table in tables
                ]
                ax.bar(
                    ind + width * j,
                    method_means,
                    width,
                    yerr=method_ses,
                    color=colors[j],
                    alpha=0.8,
                    edgecolor="white",
                )
                ci95s = [
                    all_results[dataset][method]["single_table_metrics"][base_metric][
                        table
                    ]["reference_ci"][1]
                    for table in tables
                ]

                max_value = max(max_value, max(method_means), max(ci95s))

            x_tick_width_coef = get_x_tick_width_coef(N)
            ax.set_xticks(ind + x_tick_width_coef * width)
            rotation = 20 if len(tables) > 6 else 0
            ax.set_xticklabels(tables, rotation=rotation, fontsize=fontsize)
            # y_min = 0

            if not log_scale:
                ax.set_ylim(bottom=0, top=max_value * 1.25)
            else:
                p = kwargs.get(f"{base_metric}_pow", 20)
                top = math.pow(max_value, p)
                ax.set_ylim(top=top)
            # ax.set_yticks(np.arange(y_min, 1.01, 0.1))
            ax.set_ylabel("Metric Value", fontsize=fontsize)
            ax.tick_params(axis="y", labelsize=20)

            custom_lines = [Line2D([0], [0], color=colors[i], lw=4) for i in range(N)]
            ax.legend(
                custom_lines,
                [prettify_method_name(method) for method in methods],
                loc="upper center",
                ncol=3,
                fontsize=18,
            )

            for j, table in enumerate(tables):
                ci95 = all_results[dataset][method]["single_table_metrics"][
                    base_metric
                ][table]["reference_ci"]
                # ci95 = np.mean(ci95, axis=0)
                ax.axhline(
                    y=ci95[1],
                    color="black",
                    linestyle="--",
                    linewidth=1,
                    xmin=j / len(tables),
                    xmax=(j + 1) / len(tables),
                )
                ax.axhline(
                    y=ci95[0],
                    color="black",
                    linestyle="--",
                    linewidth=1,
                    xmin=j / len(tables),
                    xmax=(j + 1) / len(tables),
                )

                if ci95[1] > ax.get_ylim()[1]:
                    y_max = ci95[1] * 1.1
                    ax.set_ylim(0, y_max)

            # set title
            if title:
                plt.title(
                    f"{base_metric_name} for dataset {prettify_dataset_name(dataset)}"
                )

            if save_figs:
                os.makedirs(save_figs_path, exist_ok=True)
                plt.savefig(f"{save_figs_path}/{dataset}_{base_metric}.png", dpi=300)


def visualize_single_table_detection_metrics_per_classifier(
    granularity_level, metric_type, all_results, datasets, methods, **kwargs
):
    for dataset in datasets:
        base_metrics, base_metric_names, save_figs, save_figs_path, methods = (
            get_dataset_info(
                granularity_level, metric_type, all_results, dataset, methods, **kwargs
            )
        )

        for base_metric, base_metric_name in zip(base_metrics, base_metric_names):
            if len(methods) == 0:
                continue

            tables = all_results[dataset][methods[0]]["single_table_metrics"][
                base_metric
            ].keys()

            N = len(methods)  # number of methods
            M = len(tables)  # number of tables

            ind = np.arange(M)
            width = 0.15

            fig, ax = plt.subplots(figsize=(10, 7))
            # set dpi
            fig.dpi = 300

            colors = plt.cm.viridis(np.linspace(0.5, 1, N))  # create a color map

            min_mean = 1
            for j, method in enumerate(methods):
                method_means = [
                    all_results[dataset][method]["single_table_metrics"][base_metric][
                        table
                    ]["accuracy"]
                    for table in tables
                ]
                min_mean = min(min_mean, min(method_means))
                method_ses = [
                    all_results[dataset][method]["single_table_metrics"][base_metric][
                        table
                    ]["SE"]
                    for table in tables
                ]
                baseline_means = np.array(
                    [
                        all_results[dataset][method]["single_table_metrics"][
                            base_metric
                        ][table]["baseline_mean"]
                        for table in tables
                    ]
                )
                baseline_ses = np.array(
                    [
                        all_results[dataset][method]["single_table_metrics"][
                            base_metric
                        ][table]["baseline_se"]
                        for table in tables
                    ]
                )
                ax.bar(
                    ind + width * j,
                    method_means,
                    width,
                    yerr=method_ses,
                    color=colors[j],
                )
                # draw a horizontal line for the baseline and standard error
                ax.hlines(
                    baseline_means,
                    ind + width * j - width / 2,
                    ind + width * j + width / 2,
                    color="k",
                )  # , linestyle='--')
                ax.hlines(
                    baseline_means + baseline_ses,
                    ind + width * j - width / 2,
                    ind + width * j + width / 2,
                    color="k",
                    linestyle="--",
                )
                ax.hlines(
                    baseline_means - baseline_ses,
                    ind + width * j - width / 2,
                    ind + width * j + width / 2,
                    color="k",
                    linestyle="--",
                )

            x_tick_width_coef = get_x_tick_width_coef(N)
            ax.set_xticks(ind + x_tick_width_coef * width)
            rotation = 20 if len(tables) > 6 else 0
            ax.set_xticklabels(tables, fontsize=10, rotation=rotation)

            y_min = 0.4 if min_mean > 0.4 else np.floor((min_mean - 0.1) * 10) / 10
            ax.set_ylim(y_min, 1.1)
            ax.set_ylabel("Metric Value")

            # Create a legend

            custom_lines = [Line2D([0], [0], color=colors[i], lw=4) for i in range(N)]
            ax.legend(custom_lines, methods, loc="upper center", ncol=N, fontsize=11)

            ax.axhline(y=0.5, color="red", linestyle="--", linewidth=1)

            # set title
            plt.title(f"{base_metric_name} for dataset {dataset}")

            if save_figs:
                os.makedirs(save_figs_path, exist_ok=True)
                plt.savefig(f"{save_figs_path}/{dataset}_{base_metric}.png", dpi=300)


def visualize_single_table_detection_metrics_per_table(
    all_results, datasets, methods, title=True, log_scale=False, fontsize=20, **kwargs
):
    for dataset in datasets:
        metrics = kwargs.get(
            "detection_metrics",
            [
                metric
                for metric in list(
                    all_results[dataset][list(all_results[dataset].keys())[0]][
                        "single_table_metrics"
                    ].keys()
                )
                if "detection" in metric.lower()
            ],
        )
        metric_names = kwargs.get("detection_metric_names", metrics)

        aggregation_metrics = kwargs.get(
            "aggregation_metrics",
            [
                metric
                for metric in list(
                    all_results[dataset][list(all_results[dataset].keys())[0]][
                        "multi_table_metrics"
                    ].keys()
                )
                if "AggregationDetection" in metric and "parent" not in metric.lower()
            ],
        )

        save_figs = kwargs.get("save_figs", False)
        save_figs_path = kwargs.get("save_figs_path", "./figs")
        save_figs_path = Path(save_figs_path) / "single_table" / "detection"

        method_order = kwargs.get(
            "method_order",
            [
                "SDV",
                "RCTGAN",
                "REALTABFORMER",
                "MOSTLYAI",
                "GRETEL_ACTGAN",
                "GRETEL_LSTM",
                "CLAVADDPM",
            ],
        )

        if method_order is not None:
            methods = [
                method
                for method in method_order
                if method in methods and method in all_results[dataset]
            ]
            methods += sorted(
                [method for method in methods if method not in method_order]
            )

        if len(methods) == 0 or len(metrics) == 0:
            continue
        method_ = list(all_results[dataset].keys())[0]
        tables = all_results[dataset][method_]["single_table_metrics"][
            list(all_results[dataset][method_]["single_table_metrics"].keys())[0]
        ].keys()
        for table in tables:
            agg_metrics = []
            if (
                aggregation_metrics
                and table
                in all_results[dataset][methods[-1]]["multi_table_metrics"][
                    aggregation_metrics[-1]
                ]
            ):
                agg_metrics = aggregation_metrics

            N = len(metrics + agg_metrics)  # number of metrics
            M = len(methods)  # number of methods
            ind = np.arange(M)  # the x locations for the groups
            width = 0.15  # the width of the bars

            fig, ax = plt.subplots(figsize=(10, 7))
            # set dpi
            fig.dpi = 300
            # make font size bigger
            # plt.rcParams.update({'font.size': 20})

            colors = plt.cm.tab20(np.linspace(0, 1, N))  # create a color map

            tab20 = plt.colormaps["tab20"]
            colors = [tab20(1), tab20(3), tab20(0), tab20(2)]

            min_mean = 1
            for j, metric in enumerate(metrics):
                metric_means = [
                    all_results[dataset][method]["single_table_metrics"][metric][table][
                        "accuracy"
                    ]
                    for method in methods
                ]
                min_mean = min(min_mean, min(metric_means))
                metric_ses = [
                    all_results[dataset][method]["single_table_metrics"][metric][table][
                        "SE"
                    ]
                    for method in methods
                ]

                ax.bar(
                    ind + width * j,
                    metric_means,
                    width,
                    yerr=metric_ses,
                    color=colors[j],
                )
                # ax.hlines(baseline_means, ind + width*j - width/2, ind + width*j + width/2, color='k')#, linestyle='--')
                # ax.hlines(baseline_means + baseline_ses, ind + width*j - width/2, ind + width*j + width/2, color='k', linestyle='--')
                # ax.hlines(baseline_means - baseline_ses, ind + width*j - width/2, ind + width*j + width/2, color='k', linestyle='--')

            for j, agg_metric in enumerate(agg_metrics):
                metric_means = [
                    all_results[dataset][method]["multi_table_metrics"][agg_metric][
                        table
                    ]["accuracy"]
                    for method in methods
                ]
                min_mean = min(min_mean, min(metric_means))
                metric_ses = [
                    all_results[dataset][method]["multi_table_metrics"][agg_metric][
                        table
                    ]["SE"]
                    for method in methods
                ]

                ax.bar(
                    ind + width * (j + len(metrics)),
                    metric_means,
                    width,
                    yerr=metric_ses,
                    color=colors[j + len(metrics)],
                )

            ax.set_ylabel("Means")
            x_tick_width_coef = get_x_tick_width_coef(N)
            ax.set_xticks(ind + x_tick_width_coef * width)
            pretty_methods = [prettify_method_name(method) for method in methods]
            ax.set_xticklabels(pretty_methods, fontsize=9.4)

            # y_min = 0.4 if min_mean > 0.4 else np.floor((min_mean - 0.1)*10)/10
            y_min = 0.3
            ax.set_ylim(y_min, 1.4)
            ax.set_yticks(np.arange(y_min, 1.01, 0.1))
            ax.set_ylabel("Classification Accuracy")

            # Create a legend
            custom_lines = [Line2D([0], [0], color=colors[i], lw=4) for i in range(N)]
            ax.legend(
                custom_lines, metric_names + agg_metrics, loc="upper left"
            )  # move the legend

            ax.axhline(y=0.5, color="red", linestyle="--", linewidth=1)

            # make figsize smaller
            fig.set_size_inches(6.5, 4)

            # set title
            if title:
                plt.title(f"Dataset {prettify_dataset_name(dataset)}, table {table}")

            if save_figs:
                os.makedirs(save_figs_path, exist_ok=True)
                plt.savefig(
                    f"{save_figs_path}/{dataset}_{table}_per_table.png", dpi=300
                )



from .denormalized_detection import DenormalizedDetection
from .aggregation_detection import (
    DenormalizedAggregationDetection,
    AggregationDetection,
    ParentChildAggregationDetection,
)
from .parent_child import ParentChildDetection



# %%
import os
from dotenv import load_dotenv
from gretel_trainer.relational import *
from gretel_trainer.relational import RelationalData
import pandas as pd
from tqdm import tqdm
from IPython.display import display, HTML
from gretel_trainer.relational import MultiTable
from gretel_client import configure_session
import argparse

# %%
# get api key from .env
load_dotenv()

configure_session(api_key=os.environ.get("GRETEL_API_KEY"), validate=True)

args = argparse.ArgumentParser()
args.add_argument("--dataset-name", type=str, default="biodegradability")
args.add_argument("--start_fold", type=int, default=0)
args.add_argument("--end_fold", type=int, default=5)
args = args.parse_args()

DATASET_NAME = args.dataset_name
METHOD_NAME='gretel'

# %%
def join_tables(parent: str, child: str, relational_data=RelationalData()):
    p_key = relational_data.get_primary_key(parent)
    f_key = ""
    for fk in relational_data.get_foreign_keys(child):
        if fk.parent_table_name == parent:
            f_key = fk.column_name
        else:
            logging.warning(
                "The input parent and child table must be related.")

    parent_df = relational_data.get_table_data(parent)
    child_df = relational_data.get_table_data(child)

    joined_data = child_df.merge(
        parent_df, how="left", left_on=p_key, right_on=f_key)

    print(f"Number of records in {child} table:\t {len(child_df)}")
    print(f"Number of records in {parent} table:\t {len(parent_df)}")
    print(f"Number of records in joined data:\t {len(joined_data)}")

    return joined_data.head()

# %%
# @title
# Alternatively, manually define relational data
# Uncomment code to run

limit = args.end_fold
for k in tqdm(range(args.start_fold, limit)):
    csv_dir = f"../../data/splits/{DATASET_NAME}/{DATASET_NAME}_leave_out_{k}"
    parent_table = (f"molecule_train_{k}", "molecule_id")
    child1_table = (f"atom_train_{k}", "atom_id")
    child1_table_fk = parent_table[1]
    child2_table = (f"bond_train_{k}", "bond_id")
    child2_table_fk = "atom_id"
    child2_table_fk2 = "atom_id2"
    tables = [
        # ("table_name", "primary_key")
        parent_table,
        child1_table,
        child2_table
    ]

    foreign_keys = [
        # ("fkey_table.fkey", "pkey_table.pkey")
        (f"{child1_table[0]}.{child1_table_fk}",
         f"{parent_table[0]}.{parent_table[1]}"),
        (f"{child2_table[0]}.{child2_table_fk}",
        f"{child1_table[0]}.{child1_table[1]}"),
        (f"{child2_table[0]}.{child2_table_fk2}",
        f"{child1_table[0]}.{child1_table[1]}"),
    ]

    relational_data = RelationalData()

    # tables_train, tables_test = utils.get_train_test_split(DATASET_NAME, k)

    for table, pk in tables:
        relational_data.add_table(
            name=table, primary_key=pk, data=tables_train[table.split("_")[0]])

    for fk, ref in foreign_keys:
        relational_data.add_foreign_key(foreign_key=fk, referencing=ref)

    # print("\033[1m Source Data: \033[0m")
    # source_data = join_tables(parent_table[0], child1_table[0], relational_data=relational_data)

    gretel_model = "lstm"
    multitable = MultiTable(
        relational_data,
        project_display_name=f"Synthesize {DATASET_NAME} - {gretel_model}",
        gretel_model=gretel_model,
        # refresh_interval=60
    )
    multitable.train()
    parent_table_name = parent_table[0].split("_")[0]
    ratio = tables_test[parent_table_name].shape[0] / tables_train[parent_table_name].shape[0]
    multitable.generate(record_size_ratio=ratio)

    synthetic_data = multitable.synthetic_output_tables
    # utils.save_data(synthetic_data, DATASET_NAME, k, method=METHOD_NAME)

# %%
# METHOD_NAME='gretel'
# synthetic_data = multitable.synthetic_output_tables
# utils.save_data(synthetic_data, DATASET_NAME, k, method=METHOD_NAME)
# %%



import re
import warnings
from typing import Union
from copy import deepcopy

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import binomtest, norm
from sklearn.pipeline import Pipeline
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import StratifiedKFold
from sdmetrics.goal import Goal
from sdmetrics.base import BaseMetric

from syntherela.utils import CustomHyperTransformer


class SingleColumnMetric(BaseMetric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def is_constant(column: pd.Series):
        constant = column.nunique() == 1
        if constant:
            warnings.warn(f"Column {column.name} is constant.")
        return constant

    @staticmethod
    def is_applicable(column_type):
        raise NotImplementedError()


class SingleTableMetric(BaseMetric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def is_applicable(metadata):
        """
        Check if the table contains at least one column that is not an id.
        """
        for column_name in metadata["columns"].keys():
            if metadata["columns"][column_name]["sdtype"] != "id":
                return True
        return False


class StatisticalBaseMetric(BaseMetric):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def validate(data):
        raise NotImplementedError()

    @staticmethod
    def compute(real_data, synthetic_data, **kwargs):
        """
        This method is used to compute the actual metric value between two samples.
        """
        raise NotImplementedError()

    def run(self, real_data, synthetic_data, **kwargs):
        """Compute this metric.

        Args:
            real_data:
                The values from the real dataset.
            synthetic_data:
                The values from the synthetic dataset.

        Returns:
            Union[float, tuple[float]]:
                Metric output or outputs.
        """
        self.validate(real_data)
        self.validate(synthetic_data)
        return self.compute(real_data, synthetic_data)


class DistanceBaseMetric(BaseMetric):
    def __init__(self, alpha=0.01, **kwargs):
        super().__init__(**kwargs)
        self.alpha = alpha

    @staticmethod
    def compute(real_data, synthetic_data, **kwargs):
        """
        This method is used to compute the actual metric value between two samples.
        """
        raise NotImplementedError()

    def run(self, real_data, synthetic_data, **kwargs):
        """Compute the reference confidence intervals using bootstrap on the real data
        and compute the matric value on real vs synthetic data."""
        reference_mean, reference_variance, reference_standard_ci = (
            self.bootstrap_reference_standard_conf_int(
                real_data, alpha=self.alpha, **kwargs
            )
        )
        bootstrap_mean, bootstrap_se = self.bootstrap_metric_estimate(
            real_data, synthetic_data, **kwargs
        )
        value = self.compute(real_data, synthetic_data, **kwargs)
        return {
            "value": value,
            "reference_mean": reference_mean,
            "reference_variance": reference_variance,
            "reference_ci": reference_standard_ci,
            "bootstrap_mean": bootstrap_mean,
            "bootstrap_se": bootstrap_se,
        }

    def boostrap_metric_values(self, data1, data2, m=100, random_state=None, **kwargs):
        # get random_state from kwargs
        if random_state is None:
            random_state = 0
        values = []
        for i in range(m):
            sample1 = data1.sample(frac=1, replace=True, random_state=random_state + i)
            sample2 = data2.sample(
                frac=1, replace=True, random_state=random_state + i + 1
            )
            # compute the metric
            val = self.compute(sample1, sample2, **kwargs)
            values.append(val)
        return values

    def bootstrap_metric_estimate(self, real_data, synthetic_data, m=1000, **kwargs):
        """Compute the bootstrap mean and standard error estimates."""
        values = self.boostrap_metric_values(real_data, synthetic_data, m=m, **kwargs)
        return np.mean(values), np.std(values) / np.sqrt(m)

    def bootstrap_reference_standard_conf_int(
        self, real_data, m=1000, alpha=0.05, **kwargs
    ):
        """Compute the standard confidence interval of the metric
        on the original data using the bootstrap method.
        """
        values = self.boostrap_metric_values(real_data, real_data, m=m, **kwargs)
        m = len(values)
        mean = np.mean(values)
        bias_adjusted_variance = np.sqrt((1 / (m - 1)) * np.sum((values - mean) ** 2))

        if self.goal == Goal.MAXIMIZE:
            z_score = norm.ppf(alpha / 2)
            conf_int = (
                mean - z_score * np.sqrt(bias_adjusted_variance),
                self.max_value,
            )
        elif self.goal == Goal.MINIMIZE:
            z_score = norm.ppf(1 - alpha)
            conf_int = (0, mean + z_score * np.sqrt(bias_adjusted_variance))
        else:
            z_score = norm.ppf(1 - alpha / 2)
            conf_int = (
                mean - z_score * np.sqrt(bias_adjusted_variance),
                mean + z_score * np.sqrt(bias_adjusted_variance),
            )

        return mean, bias_adjusted_variance, conf_int


class DetectionBaseMetric(BaseMetric):
    def __init__(
        self,
        classifier_cls,
        classifier_args={},
        random_state=None,
        folds=10,
        m=10,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.classifier_cls = classifier_cls
        self.classifier_args = classifier_args
        self.random_state = random_state
        self.folds = folds
        self.m = m
        self.classifiers = []
        self.models = []
        self.name = f"{type(self).__name__}-{classifier_cls.__name__}"

    def prepare_data(
        self,
        real_data: Union[pd.DataFrame, pd.Series],
        synthetic_data: Union[pd.DataFrame, pd.Series],
        **kwargs,
    ):
        if isinstance(real_data, pd.DataFrame):
            assert real_data.columns.equals(
                synthetic_data.columns
            ), "Columns of real and synthetic data do not match"

        # sample the same number of rows from the real and synthetic data
        n = min(len(real_data), len(synthetic_data))
        real_data = real_data.sample(n, random_state=self.random_state)
        synthetic_data = synthetic_data.sample(
            n, random_state=self.random_state + 1 if self.random_state else None
        )

        ht = CustomHyperTransformer()
        combined_data = pd.concat([real_data, synthetic_data])
        ht.fit(combined_data)
        transformed_real_data = ht.transform(real_data.copy())
        transformed_synthetic_data = ht.transform(synthetic_data.copy())
        X = pd.concat([transformed_real_data, transformed_synthetic_data])
        y = np.hstack(
            [
                np.ones(len(transformed_real_data)),
                np.zeros(len(transformed_synthetic_data)),
            ]
        )
        # replace infinite values with NaN
        X.replace([np.inf, -np.inf], np.nan, inplace=True)
        # drop constant columns
        if X.shape[1] > 1:
            X = X.loc[:, X.apply(lambda x: x.nunique() > 1)]
        return X, y

    def stratified_kfold(self, X, y, save_models=False):
        scores = []
        # Shuffle the data
        np.random.seed(self.random_state)
        idx = np.random.permutation(len(y))
        X = X.iloc[idx]
        y = y[idx]
        kf = StratifiedKFold(n_splits=self.folds)
        for i, (train_index, test_index) in enumerate(kf.split(X, y)):
            np.random.seed(self.random_state + i)
            model = Pipeline(
                [
                    ("imputer", SimpleImputer()),
                    ("scaler", StandardScaler()),
                    ("clf", self.classifier_cls(**self.classifier_args)),
                ]
            )
            model.fit(X.iloc[train_index], y[train_index])
            probs = model.predict_proba(X.iloc[test_index])
            y_pred = probs.argmax(axis=1)
            scores.extend(list((y[test_index] == y_pred).astype(int)))
            if save_models:
                self.classifiers.append(deepcopy(model["clf"]))
                self.models.append(model)
        return scores

    def compute(self, real_data, synthetic_data, metadata, **kwargs):
        X, y = self.prepare_data(real_data, synthetic_data, metadata=metadata, **kwargs)
        # save the data for feature importance methods
        self.X = X
        self.y = y
        return self.stratified_kfold(X, y, save_models=True)

    @staticmethod
    def bootstrap_sample(real_data, random_state=None, metadata=None):
        return real_data.sample(frac=1, replace=True, random_state=random_state)

    def baseline(self, real_data, metadata, m=1000, **kwargs):
        bootstrap_scores = []
        for i in range(m):
            sample1 = self.bootstrap_sample(
                real_data, random_state=i, metadata=metadata
            )
            sample2 = self.bootstrap_sample(
                real_data, random_state=i + 1, metadata=metadata
            )
            X, y = self.prepare_data(sample1, sample2, metadata=metadata, **kwargs)
            scores = self.stratified_kfold(X, y)
            bootstrap_accuracy = np.mean(scores)
            bootstrap_scores.append(bootstrap_accuracy)
        return np.mean(bootstrap_scores), np.std(bootstrap_scores) / np.sqrt(m)

    @staticmethod
    def binomial_test(x, n, p=0.5, alternative="greater"):
        """Compute the p-value of the metric using the binomial test."""
        test = binomtest(x, n, p, alternative=alternative)
        return test.statistic, test.pvalue

    def run(self, real_data, synthetic_data, metadata, **kwargs):
        """Compute this metric.

        Args:
            real_data:
                The values from the real dataset.
            synthetic_data:
                The values from the synthetic dataset.

        Returns:
            Union[float, tuple[float]]:
                Metric output or outputs.
        """
        scores = self.compute(real_data, synthetic_data, metadata=metadata, **kwargs)
        _, bin_test_p_val = self.binomial_test(
            sum(scores), len(scores), p=0.5, alternative="greater"
        )
        _, copying_p_val = self.binomial_test(
            sum(scores), len(scores), p=0.5, alternative="less"
        )
        standard_error = np.std(scores) / np.sqrt(len(scores))
        return {
            "accuracy": np.mean(scores),
            "SE": standard_error,
            "bin_test_p_val": np.round(bin_test_p_val, decimals=16),
            "copying_p_val": np.round(copying_p_val, decimals=16),
        }

    def feature_importance(self, combine_categorical=False, combine_datetime=False):
        if not len(self.classifiers):
            raise ValueError("No classifiers have been trained.")
        if not hasattr(self.classifiers[0], "feature_importances_"):
            raise ValueError(
                "The classifier does not have a feature_importances_ attribute."
            )

        features = dict()
        feature_names = self.X.columns
        for model in self.classifiers:
            for feature, importance in zip(feature_names, model.feature_importances_):
                if feature not in features:
                    features[feature] = []
                features[feature].append(importance)

        features = {k: np.array(v) for k, v in features.items()}
        if combine_categorical:
            feature_names = dict()
            for feature in features.keys():
                # check if the feature is one-hot encoded
                if not re.search("_[0-9]+$", feature):
                    continue
                feature_name = "_".join(feature.split("_")[:-1])
                if feature_name not in feature_names:
                    feature_names[feature_name] = []
                feature_names[feature_name].append(feature)
            for feature_name, feature_group in feature_names.items():
                if len(feature_group) > 1:
                    features[feature_name] = np.concatenate(
                        [features[f] for f in feature_group]
                    )
                    for f in feature_group:
                        features.pop(f)

        if combine_datetime:
            feature_names = dict()
            for feature in features.keys():
                # check if the feature is one-hot encoded
                if not (
                    feature.endswith("_Year")
                    or feature.endswith("_Month")
                    or feature.endswith("_Day")
                    or feature.endswith("_Hour")
                    or feature.endswith("_Minute")
                    or feature.endswith("_Second")
                ):
                    continue
                feature_name = "_".join(feature.split("_")[:-1])
                if feature_name not in feature_names:
                    feature_names[feature_name] = []
                feature_names[feature_name].append(feature)
            for feature_name, feature_group in feature_names.items():
                if len(feature_group) > 1:
                    features[feature_name] = np.concatenate(
                        [features[f] for f in feature_group]
                    )
                    for f in feature_group:
                        features.pop(f)

        return dict(sorted(features.items(), key=lambda x: np.mean(x[1]), reverse=True))

    def plot_feature_importance(
        self,
        metadata,
        ax=None,
        combine_categorical=False,
        combine_datetime=False,
        lab_fontsize=30,
        fontsize=23,
    ):
        features = self.feature_importance(
            combine_categorical=combine_categorical, combine_datetime=combine_datetime
        )

        def prettyify_feature_name(feature_name):
            split_name = feature_name.split("_")
            if len(split_name) > 1:
                return " ".join(
                    [
                        (
                            word.capitalize().replace("Nunique", "\#Unique")
                            if "id" not in word
                            else ""
                        )
                        for word in split_name
                    ]
                )
            return feature_name[0].upper() + feature_name[1:]

        def find_column_type(feature_name, column_info):
            for column, values in column_info.items():
                if values["sdtype"] == "id":
                    continue
                if column == feature_name:
                    return values["sdtype"]
                elif not combine_categorical and feature_name in column:
                    return values["sdtype"]
            return None

        def get_feature_type(feature_name, metadata):
            if (
                "_counts" in feature_name
                or "_mean" in feature_name
                or "_sum" in feature_name
                or "_nunique" in feature_name
            ):
                return "aggregate"

            feature_type = None
            if isinstance(metadata, dict):
                return find_column_type(feature_name, metadata["columns"])
            else:
                for table_data in metadata.to_dict()["tables"].values():
                    feature_type = find_column_type(feature_name, table_data["columns"])
                    if feature_type is not None:
                        break
            return feature_type

        colors = {
            "aggregate": "#d7191c",
            "numerical": "#fdae61",
            "datetime": "#e3d36b",
            "boolean": "#abd9e9",
            "categorical": "#2c7bb6",
        }

        if ax is None:
            _, ax = plt.subplots(figsize=(10, 10))

        for i, (feature, importance) in enumerate(features.items()):
            feature_type = get_feature_type(feature, metadata)
            y = len(features) - i - 1
            scatter = ax.scatter(
                importance,
                np.ones(len(importance)) * y,
                s=20,
                alpha=0.6,
                c=colors[feature_type],
                label=feature_type.upper(),
            )
            color = scatter.get_facecolor()[0]

            se = np.std(importance) / np.sqrt(len(importance))
            ax.errorbar(
                np.mean(importance), y, xerr=se * 1.96, c=color, capsize=3, ls="None"
            )
            ax.scatter(np.mean(importance), y, s=120, marker="v", color=color)

        xlim = ax.get_xlim()
        ax.set_xlim(0, xlim[1])
        ax.set_yticks(range(len(features)))
        pretty_feature_names = [
            prettyify_feature_name(feature) for feature in features.keys()
        ][::-1]
        ax.set_yticklabels(pretty_feature_names)
        ax.set_xlabel("Feature importance", fontsize=lab_fontsize)
        ax.tick_params(axis="x", labelsize=fontsize - 2)
        ax.tick_params(axis="y", labelsize=fontsize)
        labels = ax.get_legend_handles_labels()
        unique_labels = {label: h for h, label in zip(*labels)}
        labels_handles = [*zip(*unique_labels.items())]
        legend = labels_handles[::-1]
        ax.legend(
            *legend,
            fontsize="x-small",
            loc="lower right",
            title="Feature type",
            markerscale=2,
        )

    def plot_partial_dependence(self, feature, lab_fontsize=30, seed=None):
        features = self.feature_importance(combine_categorical=False)
        # sort by average importance
        features = {
            k: v.mean()
            for k, v in sorted(
                features.items(), key=lambda item: np.mean(item[1]), reverse=True
            )
        }
        from sklearn.inspection import PartialDependenceDisplay

        # TODO: move these functions to some utility module
        def prettyify_feature_name(feature_name):
            split_name = feature_name.split("_")
            if len(split_name) > 1:
                return " ".join(
                    [
                        (
                            word.capitalize().replace("Nunique", "\#Unique")
                            if "id" not in word
                            else ""
                        )
                        for word in split_name
                    ]
                )
            return feature_name[0].upper() + feature_name[1:]

        def get_average_pds(feature, seed=0):
            plt.ioff()
            ys = []
            disps_ind = []
            for i in range(10):
                disp_ind = PartialDependenceDisplay.from_estimator(
                    self.models[i],
                    self.X.sample(30, random_state=seed + i),
                    [feature],
                    kind="individual",
                    response_method="predict_proba",
                )
                np.random.seed(seed + i)
                disp = PartialDependenceDisplay.from_estimator(
                    self.models[i],
                    self.X,
                    [feature],
                    kind="average",
                    response_method="predict_proba",
                    subsample=0.5,
                    percentiles=(0, 1),
                )
                disps_ind.append(disp_ind)
                y = disp.axes_[0][0].lines[0].get_ydata()
                if i == 0:
                    ys.append(y)
                    x = disp.axes_[0][0].lines[0].get_xdata()

                elif i > 0 and y.shape == ys[0].shape:
                    ys.append(y)
                plt.clf()
            plt.close("all")
            return x, np.array(ys), disps_ind

        x, ys, disps = get_average_pds(feature, seed=seed)

        fig, ax = plt.subplots(figsize=(8, 6))
        disps[0].plot(ax=ax)
        for i in range(1, len(disps)):
            disps[i].plot(ax=disps[i - 1].axes_)

        ax = disps[0].axes_[0][0]
        y_mean = ys.mean(axis=0)
        y_se = ys.std(axis=0) / np.sqrt(ys.shape[0])

        ax.plot(x, y_mean, color="C0", label="Individual CEs")
        ax.plot(x, y_mean, color="C1", label="Average")
        ax.fill_between(
            x, y_mean - y_se, y_mean + y_se, alpha=0.4, color="C1", zorder=1, label="SE"
        )

        if all([x_.is_integer() for x_ in x]):
            ax.set_xticks(x)
            ax.set_xticklabels(x.astype(int))
        ax.set_xlabel(prettyify_feature_name(feature), fontsize=lab_fontsize)
        ax.set_ylabel("Partial dependence", fontsize=lab_fontsize)
        ax.legend(fontsize="xx-small", loc="lower left")



import os
import json
import argparse

from relsyndgb.metadata import Metadata
from relsyndgb.metadata import convert_metadata_to_v0

def parse_args():
    parser = argparse.ArgumentParser(description='Convert metadata to v0')
    parser.add_argument('--dataset', type=str, help='Dataset name', default = 'all')
    return parser.parse_args()

def convert_metadata_to_old_format(dataset):
    metadata = Metadata().load_from_json(f'../../data/downloads/{dataset}/metadata.json')
    meta_old = convert_metadata_to_v0(metadata)
    with open(f'../../data/downloads/{dataset}/metadata_v0.json', 'w') as f:
        json.dump(meta_old, f, indent=2)


def main():
    args = parse_args()
    dataset = args.dataset
    if dataset == 'all':
        for dataset in os.listdir('../../data/downloads'):
            if os.path.isdir(os.path.join('../../data/downloads', dataset)):
                print(f'Converting {dataset}', end='... ')
                convert_metadata_to_old_format(dataset)
                print('Done!')
    else:
        convert_metadata_to_old_format(dataset)



if __name__ == '__main__':
    main()


"""Functions for Metadata visualization."""

import warnings

import graphviz


def _get_graphviz_extension(path):
    if path:
        path_splitted = path.split('.')
        if len(path_splitted) == 1:
            raise ValueError('Path without graphviz extansion.')

        graphviz_extension = path_splitted[-1]

        if graphviz_extension not in graphviz.FORMATS:
            raise ValueError(
                '"{}" not a valid graphviz extension format.'.format(graphviz_extension)
            )

        return '.'.join(path_splitted[:-1]), graphviz_extension

    return None, None


def _add_nodes(metadata, digraph, names, details):
    """Add nodes into a `graphviz.Digraph`.

    Each node represent a metadata table.

    Args:
        metadata (Metadata):
            Metadata object to plot.
        digraph (graphviz.Digraph):
            graphviz.Digraph being built
    """
    for table in metadata.get_tables():
        if not names:
            title = ''
        elif details:
            # Append table fields
            fields = []

            for name, value in metadata.get_fields(table).items():
                if value.get('subtype') is not None:
                    fields.append('{} : {} - {}'.format(name, value['type'], value['subtype']))
                else:
                    fields.append('{} : {}'.format(name, value['type']))

            fields = r'\l'.join(fields)

            # Append table extra information
            extras = []

            primary_key = metadata.get_primary_key(table)
            if primary_key is not None:
                extras.append('Primary key: {}'.format(primary_key))

            parents = metadata.get_parents(table)
            for parent in parents:
                for foreign_key in metadata.get_foreign_keys(parent, table):
                    extras.append('Foreign key ({}): {}'.format(parent, foreign_key))

            path = metadata.get_table_meta(table).get('path')
            if path is not None:
                extras.append('Data path: {}'.format(path))

            extras = r'\l'.join(extras)

            # Add table node
            title = r'{{{}|{}\l|{}\l}}'.format(table, fields, extras)
        else:
            # Add table node only
            title = r'{{{}}}'.format(table)

        digraph.node(table, label=title)


def _add_edges(metadata, digraph, names, details):
    """Add edges into a `graphviz.Digraph`.

    Each edge represents a relationship between two metadata tables.

    Args:
        digraph (graphviz.Digraph)
    """
    for table in metadata.get_tables():
        for parent in list(metadata.get_parents(table)):
            if names and details:
                label = '\n'.join([
                    '   {}.{} > {}.{}'.format(
                        table, foreign_key,
                        parent, metadata.get_primary_key(parent)
                    )
                    for foreign_key in metadata.get_foreign_keys(parent, table)
                ])
                digraph.edge(
                    parent,
                    table,
                    label=label,
                    arrowhead='oinv'
                )
            else:
                for foreign_key in metadata.get_foreign_keys(parent, table):
                    digraph.edge(parent, table, arrowhead='oinv')


def visualize(metadata, path=None, names=True, details=True):
    """Plot metadata usign graphviz.

    Try to generate a plot using graphviz.
    If a ``path`` is provided save the output into a file.

    Args:
        metadata (Metadata):
            Metadata object to plot.
        path (str):
            Output file path to save the plot, it requires a graphviz
            supported extension. If ``None`` do not save the plot.
            Defaults to ``None``.
    """
    filename, graphviz_extension = _get_graphviz_extension(path)
    digraph = graphviz.Digraph(
        'Metadata',
        format=graphviz_extension,
        node_attr={
            "shape": "Mrecord",
            "fillcolor": "lightgoldenrod1",
            "style": "filled"
        },
    )

    _add_nodes(metadata, digraph, names, details)
    _add_edges(metadata, digraph, names, details)

    if filename:
        digraph.render(filename=filename, cleanup=True, format=graphviz_extension)
    else:
        try:
            graphviz.version()
        except graphviz.ExecutableNotFound:
            warning_message = (
                'Graphviz does not seem to be installed on this system. For full '
                'metadata visualization capabilities, please make sure to have its '
                'binaries propertly installed: https://graphviz.gitlab.io/download/'
            )
            warnings.warn(warning_message, RuntimeWarning)

        return digraph



"""Transformers module."""

import importlib
import inspect
import json
import sys
from collections import defaultdict
from copy import deepcopy
from functools import lru_cache
from pathlib import Path

from rctgan.rdt2.transformers.base import BaseTransformer
from rctgan.rdt2.transformers.boolean import BinaryEncoder
from rctgan.rdt2.transformers.categorical import FrequencyEncoder, LabelEncoder, OneHotEncoder
from rctgan.rdt2.transformers.datetime import OptimizedTimestampEncoder, UnixTimestampEncoder
from rctgan.rdt2.transformers.null import NullTransformer
from rctgan.rdt2.transformers.numerical import ClusterBasedNormalizer, FloatFormatter, GaussianNormalizer
from rctgan.rdt2.transformers.pii.anonymizer import AnonymizedFaker

__all__ = [
    'BaseTransformer',
    'BinaryEncoder',
    'ClusterBasedNormalizer',
    'FloatFormatter',
    'FrequencyEncoder',
    'GaussianNormalizer',
    'LabelEncoder',
    'NullTransformer',
    'OneHotEncoder',
    'OptimizedTimestampEncoder',
    'UnixTimestampEncoder',
    'AnonymizedFaker',
    'get_transformer_name',
    'get_transformer_class',
    'get_transformer_instance',
    'get_transformers_by_type',
    'get_default_transformers',
    'get_default_transformer',
]


def _import_addons():
    """Import all the addon modules."""
    addons_path = Path(__file__).parent / 'addons'
    for addon_json_path in addons_path.glob('*/*.json'):
        with open(addon_json_path, 'r', encoding='utf-8') as addon_json_file:
            transformers = json.load(addon_json_file).get('transformers', [])
            for transformer in transformers:
                module = transformer.rsplit('.', 1)[0]
                if module not in sys.modules:
                    importlib.import_module(module)


_import_addons()


def get_transformer_name(transformer):
    """Return the fully qualified path of the transformer.

    Args:
        transformer:
            A transformer class.

    Raises:
        ValueError:
            Crashes when the transformer is not passed as a class.

    Returns:
        string:
            The path of the transformer.
    """
    if inspect.isclass(transformer):
        return transformer.__module__ + '.' + transformer.__name__

    raise ValueError(f'The transformer {transformer} must be passed as a class.')


TRANSFORMERS = {
    get_transformer_name(transformer): transformer
    for transformer in BaseTransformer.get_subclasses()
}


DEFAULT_TRANSFORMERS = {
    'numerical': FloatFormatter(missing_value_replacement='mean'),
    'categorical': FrequencyEncoder(),
    'boolean': BinaryEncoder(missing_value_replacement='mode'),
    'datetime': UnixTimestampEncoder(missing_value_replacement='mean'),
}


def get_transformer_class(transformer):
    """Return a ``transformer`` class from a ``str``.

    Args:
        transformer (str):
            Python path.

    Returns:
        BaseTransformer:
            BaseTransformer subclass class object.
    """
    if transformer in TRANSFORMERS:
        return TRANSFORMERS[transformer]

    package, name = transformer.rsplit('.', 1)
    return getattr(importlib.import_module(package), name)


def get_transformer_instance(transformer):
    """Load a new instance of a ``Transformer``.

    The ``transformer`` is expected to be the transformers path as a ``string``,
    a transformer instance or a transformer type.

    Args:
        transformer (str or BaseTransformer):
            String with the transformer path or instance of a BaseTransformer subclass.

    Returns:
        BaseTransformer:
            BaseTransformer subclass instance.
    """
    if isinstance(transformer, BaseTransformer):
        return deepcopy(transformer)

    if inspect.isclass(transformer) and issubclass(transformer, BaseTransformer):
        return transformer()

    return get_transformer_class(transformer)()


@lru_cache()
def get_transformers_by_type():
    """Build a ``dict`` mapping sdtypes to valid existing transformers for that sdtype.

    Returns:
        dict:
            Mapping of sdtypes to a list of existing transformers that take that
            sdtype as an input.
    """
    sdtype_transformers = defaultdict(list)
    transformer_classes = BaseTransformer.get_subclasses()
    for transformer in transformer_classes:
        input_sdtype = transformer.get_input_sdtype()
        sdtype_transformers[input_sdtype].append(transformer)

    return sdtype_transformers


@lru_cache()
def get_default_transformers():
    """Build a ``dict`` mapping sdtypes to a default transformer for that sdtype.

    Returns:
        dict:
            Mapping of sdtypes to a transformer.
    """
    transformers_by_type = get_transformers_by_type()
    defaults = deepcopy(DEFAULT_TRANSFORMERS)
    for (sdtype, transformers) in transformers_by_type.items():
        if sdtype not in defaults:
            defaults[sdtype] = transformers[0]()

    return defaults


@lru_cache()
def get_default_transformer(sdtype):
    """Get default transformer for a sdtype.

    Returns:
        Transformer:
            Default transformer for sdtype.
    """
    default_transformers = get_default_transformers()
    return default_transformers[sdtype]



"""Transformers for numerical data."""
import copy
import sys
import warnings

import numpy as np
import pandas as pd
import scipy
from sklearn.mixture import BayesianGaussianMixture

from rctgan.rdt2.transformers.base import BaseTransformer
from rctgan.rdt2.transformers.null import NullTransformer

EPSILON = np.finfo(np.float32).eps
MAX_DECIMALS = sys.float_info.dig - 1


class FloatFormatter(BaseTransformer):
    """Transformer for numerical data.

    This transformer replaces integer values with their float equivalent.
    Non null float values are not modified.

    Null values are replaced using a ``NullTransformer``.

    Args:
        missing_value_replacement (object or None):
            Indicate what to do with the null values. If an integer or float is given,
            replace them with the given value. If the strings ``'mean'`` or ``'mode'`` are
            given, replace them with the corresponding aggregation. If ``None`` is given,
            do not replace them. Defaults to ``None``.
        model_missing_values (bool):
            Whether to create a new column to indicate which values were null or not. The column
            will be created only if there are null values. If ``True``, create the new column if
            there are null values. If ``False``, do not create the new column even if there
            are null values. Defaults to ``False``.
        learn_rounding_scheme (bool):
            Whether or not to learn what place to round to based on the data seen during ``fit``.
            If ``True``, the data returned by ``reverse_transform`` will be rounded to that place.
            Defaults to ``False``.
        enforce_min_max_values (bool):
            Whether or not to clip the data returned by ``reverse_transform`` to the min and
            max values seen during ``fit``. Defaults to ``False``.
    """

    INPUT_SDTYPE = 'numerical'
    DETERMINISTIC_TRANSFORM = True
    DETERMINISTIC_REVERSE = True
    COMPOSITION_IS_IDENTITY = True

    null_transformer = None
    missing_value_replacement = None
    _dtype = None
    _rounding_digits = None
    _min_value = None
    _max_value = None

    def __init__(self, missing_value_replacement=None, model_missing_values=False,
                 learn_rounding_scheme=False, enforce_min_max_values=False):
        self.missing_value_replacement = missing_value_replacement
        self.model_missing_values = model_missing_values
        self.learn_rounding_scheme = learn_rounding_scheme
        self.enforce_min_max_values = enforce_min_max_values

    def get_output_sdtypes(self):
        """Return the output sdtypes supported by the transformer.

        Returns:
            dict:
                Mapping from the transformed column names to supported sdtypes.
        """
        output_sdtypes = {
            'value': 'float',
        }
        if self.null_transformer and self.null_transformer.models_missing_values():
            output_sdtypes['is_null'] = 'float'

        return self._add_prefix(output_sdtypes)

    def is_composition_identity(self):
        """Return whether composition of transform and reverse transform produces the input data.

        Returns:
            bool:
                Whether or not transforming and then reverse transforming returns the input data.
        """
        if self.null_transformer and not self.null_transformer.models_missing_values():
            return False

        return self.COMPOSITION_IS_IDENTITY

    @staticmethod
    def _learn_rounding_digits(data):
        # check if data has any decimals
        data = np.array(data)
        roundable_data = data[~(np.isinf(data) | pd.isna(data))]
        if ((roundable_data % 1) != 0).any():
            if not (roundable_data == roundable_data.round(MAX_DECIMALS)).all():
                return None

            for decimal in range(MAX_DECIMALS + 1):
                if (roundable_data == roundable_data.round(decimal)).all():
                    return decimal

        elif len(roundable_data) > 0:
            maximum = max(abs(roundable_data))
            start = int(np.log10(maximum)) if maximum != 0 else 0
            for decimal in range(-start, 1):
                if (roundable_data == roundable_data.round(decimal)).all():
                    return decimal

        return None

    def _fit(self, data):
        """Fit the transformer to the data.

        Args:
            data (pandas.Series):
                Data to fit.
        """
        self._dtype = data.dtype
        if self.enforce_min_max_values:
            self._min_value = data.min()
            self._max_value = data.max()

        if self.learn_rounding_scheme:
            self._rounding_digits = self._learn_rounding_digits(data)

        self.null_transformer = NullTransformer(
            self.missing_value_replacement,
            self.model_missing_values
        )
        self.null_transformer.fit(data)

    def _transform(self, data):
        """Transform numerical data.

        Integer values are replaced by their float equivalent. Non null float values
        are left unmodified.

        Args:
            data (pandas.Series):
                Data to transform.

        Returns:
            numpy.ndarray
        """
        return self.null_transformer.transform(data)

    def _reverse_transform(self, data):
        """Convert data back into the original format.

        Args:
            data (pd.Series or numpy.ndarray):
                Data to transform.

        Returns:
            numpy.ndarray
        """
        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        if self.missing_value_replacement is not None:
            data = self.null_transformer.reverse_transform(data)

        if self.enforce_min_max_values:
            data = data.clip(self._min_value, self._max_value)

        is_integer = np.dtype(self._dtype).kind == 'i'
        if self.learn_rounding_scheme or is_integer:
            data = data.round(self._rounding_digits or 0)

        if pd.isna(data).any() and is_integer:
            return data

        return data.astype(self._dtype)


class GaussianNormalizer(FloatFormatter):
    r"""Transformer for numerical data based on copulas transformation.

    Transformation consists on bringing the input data to a standard normal space
    by using a combination of *cdf* and *inverse cdf* transformations:

    Given a variable :math:`x`:

    - Find the best possible marginal or use user specified one, :math:`P(x)`.
    - do :math:`u = \phi (x)` where :math:`\phi` is cumulative density function,
      given :math:`P(x)`.
    - do :math:`z = \phi_{N(0,1)}^{-1}(u)`, where :math:`\phi_{N(0,1)}^{-1}` is
      the *inverse cdf* of a *standard normal* distribution.

    The reverse transform will do the inverse of the steps above and go from :math:`z`
    to :math:`u` and then to :math:`x`.

    Args:
        model_missing_values (bool):
            Whether to create a new column to indicate which values were null or not. The column
            will be created only if there are null values. If ``True``, create the new column if
            there are null values. If ``False``, do not create the new column even if there
            are null values. Defaults to ``False``.
        learn_rounding_scheme (bool):
            Whether or not to learn what place to round to based on the data seen during ``fit``.
            If ``True``, the data returned by ``reverse_transform`` will be rounded to that place.
            Defaults to ``False``.
        enforce_min_max_values (bool):
            Whether or not to clip the data returned by ``reverse_transform`` to the min and
            max values seen during ``fit``. Defaults to ``False``.
        distribution (copulas.univariate.Univariate or str):
            Copulas univariate distribution to use. Defaults to ``parametric``. To choose from:

                * ``univariate``: Let ``copulas`` select the optimal univariate distribution.
                  This may result in non-parametric models being used.
                * ``parametric``: Let ``copulas`` select the optimal univariate distribution,
                  but restrict the selection to parametric distributions only.
                * ``bounded``: Let ``copulas`` select the optimal univariate distribution,
                  but restrict the selection to bounded distributions only.
                  This may result in non-parametric models being used.
                * ``semi_bounded``: Let ``copulas`` select the optimal univariate distribution,
                  but restrict the selection to semi-bounded distributions only.
                  This may result in non-parametric models being used.
                * ``parametric_bounded``: Let ``copulas`` select the optimal univariate
                  distribution, but restrict the selection to parametric and bounded distributions
                  only.
                * ``parametric_semi_bounded``: Let ``copulas`` select the optimal univariate
                  distribution, but restrict the selection to parametric and semi-bounded
                  distributions only.
                * ``gaussian``: Use a Gaussian distribution.
                * ``gamma``: Use a Gamma distribution.
                * ``beta``: Use a Beta distribution.
                * ``student_t``: Use a Student T distribution.
                * ``gussian_kde``: Use a GaussianKDE distribution. This model is non-parametric,
                  so using this will make ``get_parameters`` unusable.
                * ``truncated_gaussian``: Use a Truncated Gaussian distribution.
    """

    _univariate = None
    COMPOSITION_IS_IDENTITY = False

    def __init__(self, model_missing_values=False, learn_rounding_scheme=False,
                 enforce_min_max_values=False, distribution='parametric'):
        super().__init__(
            missing_value_replacement='mean',
            model_missing_values=model_missing_values,
            learn_rounding_scheme=learn_rounding_scheme,
            enforce_min_max_values=enforce_min_max_values
        )

        self.distribution = distribution  # Distribution initialized by the user

        self._distributions = self._get_distributions()
        if isinstance(distribution, str):
            distribution = self._distributions[distribution]

        self._distribution = distribution

    @staticmethod
    def _get_distributions():
        try:
            from copulas import univariate  # pylint: disable=import-outside-toplevel
        except ImportError as error:
            error.msg += (
                '\n\nIt seems like `copulas` is not installed.\n'
                'Please install it using:\n\n    pip install rdt2[copulas]'
            )
            raise

        return {
            'univariate': univariate.Univariate,
            'parametric': (
                univariate.Univariate, {
                    'parametric': univariate.ParametricType.PARAMETRIC,
                },
            ),
            'bounded': (
                univariate.Univariate,
                {
                    'bounded': univariate.BoundedType.BOUNDED,
                },
            ),
            'semi_bounded': (
                univariate.Univariate,
                {
                    'bounded': univariate.BoundedType.SEMI_BOUNDED,
                },
            ),
            'parametric_bounded': (
                univariate.Univariate,
                {
                    'parametric': univariate.ParametricType.PARAMETRIC,
                    'bounded': univariate.BoundedType.BOUNDED,
                },
            ),
            'parametric_semi_bounded': (
                univariate.Univariate,
                {
                    'parametric': univariate.ParametricType.PARAMETRIC,
                    'bounded': univariate.BoundedType.SEMI_BOUNDED,
                },
            ),
            'gaussian': univariate.GaussianUnivariate,
            'gamma': univariate.GammaUnivariate,
            'beta': univariate.BetaUnivariate,
            'student_t': univariate.StudentTUnivariate,
            'gaussian_kde': univariate.GaussianKDE,
            'truncated_gaussian': univariate.TruncatedGaussian,
        }

    def _get_univariate(self):
        distribution = self._distribution
        if isinstance(distribution, self._distributions['univariate']):
            return copy.deepcopy(distribution)
        if isinstance(distribution, tuple):
            return distribution[0](**distribution[1])
        if isinstance(distribution, type) and \
           issubclass(distribution, self._distributions['univariate']):
            return distribution()

        raise TypeError(f'Invalid distribution: {distribution}')

    def _fit(self, data):
        """Fit the transformer to the data.

        Args:
            data (pandas.Series):
                Data to fit to.
        """
        self._univariate = self._get_univariate()

        super()._fit(data)
        data = super()._transform(data)
        if data.ndim > 1:
            data = data[:, 0]

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self._univariate.fit(data)

    def _copula_transform(self, data):
        cdf = self._univariate.cdf(data)
        return scipy.stats.norm.ppf(cdf.clip(0 + EPSILON, 1 - EPSILON))

    def _transform(self, data):
        """Transform numerical data.

        Args:
            data (pandas.Series):
                Data to transform.

        Returns:
            numpy.ndarray
        """
        transformed = super()._transform(data)
        if transformed.ndim > 1:
            transformed[:, 0] = self._copula_transform(transformed[:, 0])
        else:
            transformed = self._copula_transform(transformed)

        return transformed

    def _reverse_transform(self, data):
        """Convert data back into the original format.

        Args:
            data (pd.Series or numpy.ndarray):
                Data to transform.

        Returns:
            pandas.Series
        """
        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        if data.ndim > 1:
            data[:, 0] = self._univariate.ppf(scipy.stats.norm.cdf(data[:, 0]))
        else:
            data = self._univariate.ppf(scipy.stats.norm.cdf(data))

        return super()._reverse_transform(data)


class ClusterBasedNormalizer(FloatFormatter):
    """Transformer for numerical data using a Bayesian Gaussian Mixture Model.

    This transformation takes a numerical value and transforms it using a Bayesian GMM
    model. It generates two outputs, a discrete value which indicates the selected
    'component' of the GMM and a continuous value which represents the normalized value
    based on the mean and std of the selected component.

    Args:
        model_missing_values (bool):
            Whether to create a new column to indicate which values were null or not. The column
            will be created only if there are null values. If ``True``, create the new column if
            there are null values. If ``False``, do not create the new column even if there
            are null values. Defaults to ``False``.
        learn_rounding_scheme (bool):
            Whether or not to learn what place to round to based on the data seen during ``fit``.
            If ``True``, the data returned by ``reverse_transform`` will be rounded to that place.
            Defaults to ``False``.
        enforce_min_max_values (bool):
            Whether or not to clip the data returned by ``reverse_transform`` to the min and
            max values seen during ``fit``. Defaults to ``False``.
        max_clusters (int):
            The maximum number of mixture components. Depending on the data, the model may select
            fewer components (based on the ``weight_threshold``).
            Defaults to 10.
        weight_threshold (int, float):
            The minimum value a component weight can take to be considered a valid component.
            ``weights_`` under this value will be ignored.
            Defaults to 0.005.

    Attributes:
        _bgm_transformer:
            An instance of sklearn`s ``BayesianGaussianMixture`` class.
        valid_component_indicator:
            An array indicating the valid components. If the weight of a component is greater
            than the ``weight_threshold``, it's indicated with True, otherwise it's set to False.
    """

    STD_MULTIPLIER = 4
    DETERMINISTIC_TRANSFORM = False
    DETERMINISTIC_REVERSE = True
    COMPOSITION_IS_IDENTITY = False

    _bgm_transformer = None
    valid_component_indicator = None

    def __init__(self, model_missing_values=False, learn_rounding_scheme=False,
                 enforce_min_max_values=False, max_clusters=10, weight_threshold=0.005):
        super().__init__(
            missing_value_replacement='mean',
            model_missing_values=model_missing_values,
            learn_rounding_scheme=learn_rounding_scheme,
            enforce_min_max_values=enforce_min_max_values
        )
        self.max_clusters = max_clusters
        self.weight_threshold = weight_threshold

    def get_output_sdtypes(self):
        """Return the output sdtypes supported by the transformer.

        Returns:
            dict:
                Mapping from the transformed column names to supported sdtypes.
        """
        output_sdtypes = {
            'normalized': 'float',
            'component': 'categorical'
        }
        if self.null_transformer and self.null_transformer.models_missing_values():
            output_sdtypes['is_null'] = 'float'

        return self._add_prefix(output_sdtypes)

    def _fit(self, data):
        """Fit the transformer to the data.

        Args:
            data (pandas.Series):
                Data to fit to.
        """
        self._bgm_transformer = BayesianGaussianMixture(
            n_components=self.max_clusters,
            weight_concentration_prior_type='dirichlet_process',
            weight_concentration_prior=0.001,
            n_init=1
        )

        super()._fit(data)
        data = super()._transform(data)
        if data.ndim > 1:
            data = data[:, 0]

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            self._bgm_transformer.fit(data.reshape(-1, 1))

        self.valid_component_indicator = self._bgm_transformer.weights_ > self.weight_threshold

    def _transform(self, data):
        """Transform the numerical data.

        Args:
            data (pandas.Series):
                Data to transform.

        Returns:
            numpy.ndarray.
        """
        data = super()._transform(data)
        if data.ndim > 1:
            data, model_missing_values = data[:, 0], data[:, 1]

        data = data.reshape((len(data), 1))
        means = self._bgm_transformer.means_.reshape((1, self.max_clusters))

        stds = np.sqrt(self._bgm_transformer.covariances_).reshape((1, self.max_clusters))
        normalized_values = (data - means) / (self.STD_MULTIPLIER * stds)
        normalized_values = normalized_values[:, self.valid_component_indicator]
        component_probs = self._bgm_transformer.predict_proba(data)
        component_probs = component_probs[:, self.valid_component_indicator]

        selected_component = np.zeros(len(data), dtype='int')
        for i in range(len(data)):
            component_prob_t = component_probs[i] + 1e-6
            component_prob_t = component_prob_t / component_prob_t.sum()
            selected_component[i] = np.random.choice(
                np.arange(self.valid_component_indicator.sum()),
                p=component_prob_t
            )

        aranged = np.arange(len(data))
        normalized = normalized_values[aranged, selected_component].reshape([-1, 1])
        normalized = np.clip(normalized, -.99, .99)
        normalized = normalized[:, 0]
        rows = [normalized, selected_component]
        if self.null_transformer and self.null_transformer.models_missing_values():
            rows.append(model_missing_values)

        return np.stack(rows, axis=1)  # noqa: PD013

    def _reverse_transform_helper(self, data):
        normalized = np.clip(data[:, 0], -1, 1)
        means = self._bgm_transformer.means_.reshape([-1])
        stds = np.sqrt(self._bgm_transformer.covariances_).reshape([-1])
        selected_component = data[:, 1].astype(int)

        std_t = stds[self.valid_component_indicator][selected_component]
        mean_t = means[self.valid_component_indicator][selected_component]
        reversed_data = normalized * self.STD_MULTIPLIER * std_t + mean_t

        return reversed_data

    def _reverse_transform(self, data):
        """Convert data back into the original format.

        Args:
            data (pd.DataFrame or numpy.ndarray):
                Data to transform.

        Returns:
            pandas.Series.
        """
        if not isinstance(data, np.ndarray):
            data = data.to_numpy()

        recovered_data = self._reverse_transform_helper(data)
        if self.null_transformer and self.null_transformer.models_missing_values():
            data = np.stack([recovered_data, data[:, -1]], axis=1)  # noqa: PD013
        else:
            data = recovered_data

        return super()._reverse_transform(data)



"""Dataset Generators for categorical transformers."""

from abc import ABC

import numpy as np

from rctgan.rdt2.performance.datasets.base import BaseDatasetGenerator
from rctgan.rdt2.performance.datasets.datetime import RandomGapDatetimeGenerator
from rctgan.rdt2.performance.datasets.utils import add_nans


class CategoricalGenerator(BaseDatasetGenerator, ABC):
    """Base class for generators that generate catgorical data."""

    SDTYPE = 'categorical'


class RandomIntegerGenerator(CategoricalGenerator):
    """Generator that creates an array of random integers."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        categories = [1, 2, 3, 4, 5]
        return np.random.choice(a=categories, size=num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 400.0
            },
            'transform': {
                'time': 5e-06,
                'memory': 400.0
            },
            'reverse_transform': {
                'time': 1e-05,
                'memory': 1000.0,
            }
        }


class RandomIntegerNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of random integers with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(RandomIntegerGenerator.generate(num_rows).astype(float))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 400.0
            },
            'transform': {
                'time': 5e-06,
                'memory': 1000.0
            },
            'reverse_transform': {
                'time': 1e-05,
                'memory': 1000.0,
            }
        }


class RandomStringGenerator(CategoricalGenerator):
    """Generator that creates an array of random strings."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        categories = ['Alice', 'Bob', 'Charlie', 'Dave', 'Eve']
        return np.random.choice(a=categories, size=num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 500.0
            },
            'transform': {
                'time': 1e-05,
                'memory': 500.0
            },
            'reverse_transform': {
                'time': 1e-05,
                'memory': 1000.0,
            }
        }


class RandomStringNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of random strings with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(RandomStringGenerator.generate(num_rows).astype('O'))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 400.0
            },
            'transform': {
                'time': 1e-05,
                'memory': 1000.0
            },
            'reverse_transform': {
                'time': 1e-05,
                'memory': 1000.0,
            }
        }


class RandomMixedGenerator(CategoricalGenerator):
    """Generator that creates an array of random mixed types.

    Mixed types include: int, float, bool, string, datetime.
    """

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        cat_size = 5
        categories = np.hstack([
            cat.astype('O') for cat in [
                RandomGapDatetimeGenerator.generate(cat_size),
                np.random.randint(0, 100, cat_size),
                np.random.uniform(0, 100, cat_size),
                np.arange(cat_size).astype(str),
                np.array([True, False])
            ]
        ])

        return np.random.choice(a=categories, size=num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 400.0
            },
            'transform': {
                'time': 1e-05,
                'memory': 1000.0
            },
            'reverse_transform': {
                'time': 1e-05,
                'memory': 2000.0,
            }
        }


class RandomMixedNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of random mixed types with nans.

    Mixed types include: int, float, bool, string, datetime.
    """

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        array = RandomMixedGenerator.generate(num_rows)

        length = len(array)
        num_nulls = np.random.randint(1, length)
        nulls_idx = np.random.choice(range(length), num_nulls)
        nulls = np.random.choice([np.nan, float('nan'), None], num_nulls)
        array[nulls_idx] = nulls

        return array

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 400.0
            },
            'transform': {
                'time': 1e-05,
                'memory': 2000.0
            },
            'reverse_transform': {
                'time': 1e-05,
                'memory': 2000.0,
            }
        }


class SingleIntegerGenerator(CategoricalGenerator):
    """Generator that creates an array with a single integer."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        constant = np.random.randint(0, 100)
        return np.full(num_rows, constant)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 400.0
            },
            'transform': {
                'time': 3e-06,
                'memory': 200.0
            },
            'reverse_transform': {
                'time': 1e-05,
                'memory': 400.0,
            }
        }


class SingleIntegerNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array with a single integer with some nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(SingleIntegerGenerator.generate(num_rows).astype(float))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 400.0
            },
            'transform': {
                'time': 3e-06,
                'memory': 200.0
            },
            'reverse_transform': {
                'time': 1e-05,
                'memory': 500.0,
            }
        }


class SingleStringGenerator(CategoricalGenerator):
    """Generator that creates an array of a single string."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        constant = 'A'
        return np.full(num_rows, constant)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 400.0
            },
            'transform': {
                'time': 4e-06,
                'memory': 200.0
            },
            'reverse_transform': {
                'time': 1e-05,
                'memory': 400.0,
            }
        }


class SingleStringNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of a single string with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(SingleStringGenerator.generate(num_rows).astype('O'))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 400.0
            },
            'transform': {
                'time': 3e-06,
                'memory': 200.0
            },
            'reverse_transform': {
                'time': 1e-05,
                'memory': 500.0,
            }
        }


class UniqueIntegerGenerator(CategoricalGenerator):
    """Generator that creates an array of unique integers."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return np.arange(num_rows)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 2e-05,
                'memory': 2000.0
            },
            'transform': {
                'time': 0.0002,
                'memory': 500000.0
            },
            'reverse_transform': {
                'time': 0.0003,
                'memory': 1000000.0,
            }
        }


class UniqueIntegerNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of unique integers with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(UniqueIntegerGenerator.generate(num_rows))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 1e-05,
                'memory': 1000.0
            },
            'transform': {
                'time': 0.0002,
                'memory': 1000000.0
            },
            'reverse_transform': {
                'time': 0.0002,
                'memory': 1000000.0,
            }
        }


class UniqueStringGenerator(CategoricalGenerator):
    """Generator that creates an array of unique strings."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return np.arange(num_rows).astype(str)

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 2e-05,
                'memory': 2000.0
            },
            'transform': {
                'time': 0.0002,
                'memory': 500000.0
            },
            'reverse_transform': {
                'time': 0.0003,
                'memory': 1000000.0,
            }
        }


class UniqueStringNaNsGenerator(CategoricalGenerator):
    """Generator that creates an array of unique strings with nans."""

    @staticmethod
    def generate(num_rows):
        """Generate a ``num_rows`` number of rows."""
        return add_nans(UniqueStringGenerator.generate(num_rows).astype('O'))

    @staticmethod
    def get_performance_thresholds():
        """Return the expected threseholds."""
        return {
            'fit': {
                'time': 2e-05,
                'memory': 1000.0
            },
            'transform': {
                'time': 0.0005,
                'memory': 1000000.0
            },
            'reverse_transform': {
                'time': 0.0002,
                'memory': 1000000.0,
            }
        }



from typing import Union
from copy import deepcopy

from syntherela.metadata import drop_ids, Metadata
from syntherela.data import denormalize_tables, make_column_names_unique
from syntherela.metrics.base import DetectionBaseMetric


class DenormalizedDetection(DetectionBaseMetric):
    @staticmethod
    def bootstrap_sample(
        real_data: dict, metadata: Metadata, random_state: Union[int, None] = None
    ) -> dict:
        bootstrapped_tables = dict()
        root_tables = metadata.get_root_tables()

        for table in root_tables:
            primary_key = metadata.get_primary_key(table)
            bootstrapped_tables[table] = real_data[table].sample(
                frac=1, replace=True, random_state=random_state
            )
            bootstrapped_tables[table][f"{primary_key}_duplicated"] = (
                bootstrapped_tables[table][primary_key].duplicated()
            )
        for relationship in metadata.relationships:
            child_table = relationship["child_table_name"]
            child_fk = relationship["child_foreign_key"]
            parent_table = relationship["parent_table_name"]
            parent_pk = relationship["parent_primary_key"]
            if child_table in bootstrapped_tables:
                child_table_pk = metadata.get_primary_key(child_table)
                new_data = real_data[child_table][
                    real_data[child_table][child_fk].isin(
                        bootstrapped_tables[parent_table][parent_pk]
                    )
                ]
                new_data.set_index(child_table_pk, inplace=True)
                bootstrapped_tables[child_table].set_index(child_table_pk, inplace=True)
                bootstrapped_tables[child_table] = (
                    bootstrapped_tables[child_table]
                    .combine_first(new_data)
                    .reset_index()
                )
            else:
                bootstrapped_tables[child_table] = real_data[child_table][
                    real_data[child_table][child_fk].isin(
                        bootstrapped_tables[parent_table][parent_pk]
                    )
                ]

        return bootstrapped_tables

    def prepare_data(self, real_data, synthetic_data, metadata):
        real_data_unique, synthetic_data_unique, metadata_unique = (
            make_column_names_unique(
                real_data.copy(),
                synthetic_data.copy(),
                deepcopy(metadata),
                validate=False,
            )
        )
        denormalized_real_data = denormalize_tables(real_data_unique, metadata_unique)
        denormalized_synthetic_data = denormalize_tables(
            synthetic_data_unique, metadata_unique
        )
        for table in metadata_unique.get_tables():
            table_metadata = metadata_unique.tables[table].to_dict()
            denormalized_real_data = drop_ids(denormalized_real_data, table_metadata)
            denormalized_synthetic_data = drop_ids(
                denormalized_synthetic_data, table_metadata
            )
        return super().prepare_data(denormalized_real_data, denormalized_synthetic_data)



from .hellinger_distance import HellingerDistance
from .jensen_shannon_distance import JensenShannonDistance
from .wasserstein_distance import WassersteinDistance
from .total_variation_distance import TotalVariationDistance



import numpy as np
import pandas as pd
from sdmetrics.goal import Goal
from sdmetrics.utils import is_datetime

from syntherela.metadata import drop_ids
from syntherela.metrics.base import DistanceBaseMetric, SingleTableMetric


class PairwiseCorrelationDifference(DistanceBaseMetric, SingleTableMetric):
    def __init__(self, norm_order="fro", correlation_method="pearson", **kwargs):
        super().__init__(**kwargs)
        self.name = "PairwiseCorrelationDifference"
        self.goal = Goal.MINIMIZE
        self.norm_order = norm_order
        self.correlation_method = correlation_method
        self.min_value = 0.0
        self.max_value = 1.0

    @staticmethod
    def is_applicable(metadata):
        """
        Check if the table contains at least one column that is not an id.
        """
        numeric_count = 0
        for column_name in metadata["columns"].keys():
            if (
                metadata["columns"][column_name]["sdtype"] == "numerical"
                or metadata["columns"][column_name]["sdtype"] == "datetime"
            ):
                numeric_count += 1
        return numeric_count > 1

    def compute(self, original_table, sythetic_table, metadata, **kwargs):
        """
        Based on:
        Andre Goncalves, Priyadip Ray, Braden Soper, Jennifer Stevens, Linda Coyle & Ana Paula Sales (2020).
        Generation and evaluation of synthetic patient data.
        https://bmcmedresmethodol.biomedcentral.com/articles/10.1186/s12874-020-00977-1
        """
        orig = original_table.copy()
        synth = sythetic_table.copy()

        orig = drop_ids(orig, metadata)
        synth = drop_ids(synth, metadata)

        zero_variance_columns = []
        for col in orig.columns:
            if orig[col].dtype.name in ("object", "category"):
                orig.drop(col, axis=1, inplace=True)
                synth.drop(col, axis=1, inplace=True)
                continue
            elif is_datetime(orig[col]):
                orig[col] = pd.to_numeric(
                    orig[col], errors="coerce", downcast="integer"
                )
                synth[col] = pd.to_numeric(
                    synth[col], errors="coerce", downcast="integer"
                )

        # drop nan values
        orig.dropna(inplace=True)
        synth.dropna(inplace=True)

        # drop columns with zero variance
        std_orig = orig.std()
        std_synth = synth.std()
        zero_variance_columns = set(
            std_orig[std_orig == 0].index.tolist()
            + std_synth[std_synth == 0].index.tolist()
        )
        orig.drop(columns=zero_variance_columns, inplace=True)
        synth.drop(columns=zero_variance_columns, inplace=True)

        # compute the correlation matrix
        orig_corr = orig.corr(method=self.correlation_method)
        synth_corr = synth.corr(method=self.correlation_method)

        return np.linalg.norm(orig_corr - synth_corr, ord=self.norm_order).astype(float)



import json

import numpy as np
import pandas as pd
from sdmetrics.utils import HyperTransformer
from sklearn.preprocessing import OneHotEncoder


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.bool_):
            return bool(obj)
        return super(NpEncoder, self).default(obj)


class CustomHyperTransformer(HyperTransformer):
    """
    CustomHyperTransformer overrides the transform method of HyperTransformer
    so that the feature names are preserved for one-hot-encoded columns.
    """

    def fit(self, data):
        """Fit the HyperTransformer to the given data.

        Args:
            data (pandas.DataFrame):
                The data to transform.
        """
        if not isinstance(data, pd.DataFrame):
            data = pd.DataFrame(data)

        for field in data:
            kind = data[field].dropna().infer_objects().dtype.kind
            self.column_kind[field] = kind

            if kind == "i" or kind == "f":
                # Numerical column.
                self.column_transforms[field] = {"mean": data[field].mean()}
            elif kind == "b":
                # Boolean column.
                numeric = pd.to_numeric(data[field], errors="coerce").astype(float)
                self.column_transforms[field] = {"mode": numeric.mode().iloc[0]}
            elif kind == "O":
                # Categorical column.
                col_data = pd.DataFrame({"field": data[field]})
                enc = OneHotEncoder(handle_unknown="ignore")
                enc.fit(col_data)
                self.column_transforms[field] = {"one_hot_encoder": enc}
            elif kind == "M":
                # Datetime column.
                nulls = data[field].isna()
                dates = pd.to_datetime(data[field][~nulls], errors="coerce")
                has_hours = dates.dt.hour.sum() > 0
                has_minutes = dates.dt.minute.sum() > 0
                has_seconds = dates.dt.second.sum() > 0
                has_microseconds = dates.dt.microsecond.sum() > 0
                self.column_transforms[field] = {
                    "has_hours": has_hours,
                    "has_minutes": has_minutes,
                    "has_seconds": has_seconds,
                    "has_microseconds": has_microseconds,
                }

    def transform(self, data):
        """Transform the given data based on the data type of each column.

        Args:
            data (pandas.DataFrame):
                The data to transform.

        Returns:
            pandas.DataFrame:
                The transformed data.
        """
        if not isinstance(data, pd.DataFrame):
            data = pd.DataFrame(data)

        for field in data:
            transform_info = self.column_transforms[field]

            kind = self.column_kind[field]
            if kind == "i" or kind == "f":
                # Numerical column.
                data[field] = data[field].fillna(transform_info["mean"])
            elif kind == "b":
                # Boolean column.
                data[field] = pd.to_numeric(data[field], errors="coerce").astype(float)
                data[field] = data[field].fillna(transform_info["mode"])
            elif kind == "O":
                # Categorical column.
                col_data = pd.DataFrame({"field": data[field]})
                out = transform_info["one_hot_encoder"].transform(col_data).toarray()
                transformed = pd.DataFrame(
                    out, columns=[f"{field}_{i}" for i in range(np.shape(out)[1])]
                )
                data = data.drop(columns=[field])
                data = pd.concat([data, transformed.set_index(data.index)], axis=1)
            elif kind == "M":
                # Datetime column.
                nulls = data[field].isnull()
                data[field] = pd.to_datetime(data[field], errors="coerce")
                data[f"{field}_Year"] = data[field].dt.year
                data[f"{field}_Month"] = data[field].dt.month
                data[f"{field}_Day"] = data[field].dt.day
                data.loc[nulls, f"{field}_Year"] = np.nan
                data.loc[nulls, f"{field}_Month"] = np.nan
                data.loc[nulls, f"{field}_Day"] = np.nan
                if transform_info["has_hours"]:
                    data[f"{field}_Hour"] = data[field].dt.hour
                    data.loc[nulls, f"{field}_Hour"] = np.nan
                if transform_info["has_minutes"]:
                    data[f"{field}_Minute"] = data[field].dt.minute
                    data.loc[nulls, f"{field}_Minute"] = np.nan
                if transform_info["has_seconds"]:
                    data[f"{field}_Second"] = data[field].dt.second
                    data.loc[nulls, f"{field}_Second"] = np.nan
                if transform_info["has_microseconds"]:
                    data[f"{field}_Microsecond"] = data[field].dt.microsecond
                    data.loc[nulls, f"{field}_Microsecond"] = np.nan
                data = data.drop(columns=[field])
        return data



from syntherela.metrics.base import DetectionBaseMetric, SingleColumnMetric


class SingleColumnDetection(DetectionBaseMetric, SingleColumnMetric):
    @staticmethod
    def is_applicable(column_type):
        return column_type in ["categorical", "datetime", "numerical", "boolean"]



import shutil
from syntherela.benchmark import Benchmark


benchmark = Benchmark(
    real_data_dir="data/original",
    synthetic_data_dir="data/synthetic",
    results_dir="results/1",
    benchmark_name="Benchmark",
    run_id="1",
    sample_id="sample1",
    datasets=["Biodegradability_v1"],
    methods=["SDV", "RCTGAN", "MOSTLYAI", "GRETEL_ACTGAN", "GRETEL_LSTM"],
    validate_metadata=False,
)

benchmark.read_results()

benchmark.visualize_single_table_metrics(
    save_figs=True,
    save_figs_path="results/figures/tmp/",
    distance=False,
    detection=True,
    title=False,
)

shutil.copy(
    "results/figures/tmp/single_table/detection/Biodegradability_v1_molecule_per_table.png",
    "results/figures/figure3a.png",
)

benchmark = Benchmark(
    real_data_dir="data/original",
    synthetic_data_dir="data/synthetic",
    results_dir="results/1",
    benchmark_name="Benchmark",
    run_id="1",
    sample_id="sample1",
    datasets=["rossmann_subsampled"],
    methods=[
        "SDV",
        "RCTGAN",
        "REALTABFORMER",
        "MOSTLYAI",
        "GRETEL_ACTGAN",
        "GRETEL_LSTM",
        "CLAVADDPM",
    ],
    validate_metadata=False,
)

benchmark.read_results()

benchmark.visualize_single_table_metrics(
    save_figs=True,
    save_figs_path="results/figures/tmp/",
    distance=False,
    detection=True,
    title=False,
)

shutil.copy(
    "results/figures/tmp/single_table/detection/rossmann_subsampled_store_per_table.png",
    "results/figures/figure3b.png",
)
# remove the tmp folder
shutil.rmtree("results/figures/tmp")
